from typing import Optional, Union, List

from quickstats import AbstractObject
from quickstats.components import ExtendedModel, ExtendedMinimizer
from quickstats.components.basics import WSArgument
from quickstats.utils.common_utils import get_class_that_defined_method
from quickstats.interface.root import RooArgSet
import ROOT

class AnalysisObject(AbstractObject):
    
    kInitialSnapshotName = "initialSnapshot"
    kCurrentSnapshotName = "currentSnapshot"
    kTempSnapshotName    = "tmpSnapshot"
    
    def __init__(self, filename:Optional[str]=None, poi_name:Optional[Union[str, List[str]]]=None,
                 data_name:str='combData', binned_likelihood:bool=True,
                 fix_param:str='', profile_param:str='', ws_name:Optional[str]=None, 
                 mc_name:Optional[str]=None, snapshot_name:Optional[Union[List[str], str]]=None,
                 minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', precision:float=0.001, 
                 eps:float=1.0, retry:int=1, strategy:int=1, print_level:int=-1, timer:bool=False,
                 num_cpu:int=1, offset:bool=True, optimize:int=2, eigen:bool=False, hesse:bool=False,
                 improve:int=0, fix_cache:bool=True, fix_multi:bool=True, max_calls:int=-1, 
                 max_iters:int=-1, constrain_nuis:bool=True, batch_mode:bool=False,
                 int_bin_precision:float=-1., preset_param:bool=False, minimizer_offset:int=1,
                 minimizer_cls=None, verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(verbosity=verbosity)    
        self.model = None
        self.minimizer = None
        self._use_blind_range = False
        if minimizer_cls is None:
            self.minimizer_cls = ExtendedMinimizer
        else:
            self.minimizer_cls = minimizer_cls
        if filename is not None:
            model_options = {
                "fname"             : filename,
                "ws_name"           : ws_name,
                "mc_name"           : mc_name,
                "data_name"         : data_name,
                "snapshot_name"     : snapshot_name,
                "binned_likelihood" : binned_likelihood,
                "fix_cache"         : fix_cache,
                "fix_multi"         : fix_multi
            }
            self.setup_model(**model_options)
            self.model._load_floating_auxiliary_variables()
            self.save_snapshot(self.kInitialSnapshotName, WSArgument.MUTABLE)
            self.set_poi(poi_name)
            if preset_param:
                self.preset_parameters()
            self.setup_parameters(fix_param, profile_param, update_snapshot=True)
            minimizer_options = {
                "minimizer_type"    : minimizer_type,
                "minimizer_algo"    : minimizer_algo,
                "precision"         : precision,
                "eps"               : eps,
                "retry"             : retry,
                "strategy"          : strategy,
                "num_cpu"           : num_cpu,
                "offset"            : offset,
                "minimizer_offset"  : minimizer_offset,
                "optimize"          : optimize,
                "eigen"             : eigen,
                "hesse"             : hesse,
                "improve"           : improve,
                "max_calls"         : max_calls,
                "max_iters"         : max_iters,
                "print_level"       : print_level,
                "timer"             : timer,
                "constrain_nuis"    : constrain_nuis,
                "batch_mode"        : batch_mode,
                "int_bin_precision" : int_bin_precision,
                "verbosity"         : verbosity
            }
            self.setup_minimizer(**minimizer_options)
            self.sanity_check()
        
    def sanity_check(self):
        aux_vars = self.model.get_variables("auxiliary")
        floating_aux_vars = [v.GetName() for v in aux_vars if not v.isConstant()]
        if len(floating_aux_vars) > 0:
            self.stdout.warning("WARNING: The following auxiliary variables (variables that are not "
                                "part of the POIs, observables, nuisance parameters and global "
                                f"observables) are floating: {','.join(floating_aux_vars)}. If this is "
                                "not intended, please make sure to fix them before fitting.", "red")            
    
    def preset_parameters(self):
        ROOT.RooStats.SetAllConstant(self.model.global_observables, True)
        #ROOT.RooStats.SetAllConstant(self.model.nuisance_parameters, False)
        ROOT.RooStats.SetAllConstant(self.model.pois, True)
        self.stdout.info("INFO: Preset POIs as constant parameters.")
        #self.stdout.info("INFO: Preset nuisance parameters as floating parameters.")
        self.stdout.info("INFO: Preset global observables as constant parameters.")
    
    @property
    def use_blind_range(self):
        return self._use_blind_range
    
    @property
    def minimizer_options(self):
        return self.minimizer.config
    
    @property
    def nll_commands(self):
        return self.minimizer.nll_commands
    
    # black magic
    def _inherit_init(self, init_func, **kwargs):
        import inspect
        this_parameters = list(inspect.signature(AnalysisObject.__init__).parameters)
        if "self" in this_parameters:
            this_parameters.remove("self")
        that_parameters = list(inspect.signature(init_func).parameters)
        is_calling_this = set(this_parameters) == set(that_parameters)
        if is_calling_this:
            init_func(**kwargs)
        else:
            that_kwargs = {k:v for k,v in kwargs.items() if k in that_parameters}
            this_kwargs = {k:v for k,v in kwargs.items() if k not in that_parameters}
            init_func(config=this_kwargs, **that_kwargs)

    def set_poi(self, poi_name:Optional[Union[str, List[str]]]=None):
        # multi POI case
        if isinstance(poi_name, list):
            self.poi = ROOT.RooArgSet()
            true_poi_names = []
            # remove duplicates
            poi_name = list(set(poi_name))
            for pname in poi_name:
                poi = self.model.get_poi(pname)
                self.poi.add(poi)
                true_poi_names.append(poi.GetName())
            if true_poi_names:
                self.stdout.info('INFO: POIs set to {}'.format(", ".join([f"\"{name}\"" for name in true_poi_names])))
        else:
            self.poi = self.model.get_poi(poi_name)
            if poi_name is not None:
                self.stdout.info(f'INFO: POI set to "{poi_name}"')
        
    def setup_model(self, **kwargs):
        model = ExtendedModel(**kwargs, verbosity=self.stdout.verbosity)
        self.model = model
    
    def setup_parameters(self, fix_param:str='', profile_param:str='', update_snapshot:bool=True):
        if not self.model:
            raise RuntimeError('uninitialized analysis object')
        fixed_parameters = []
        profiled_parameters = []
        if fix_param:
            fixed_parameters = self.model.fix_parameters(fix_param)
        if profile_param:
            profiled_parameters = self.model.profile_parameters(profile_param)
        self._update_floating_auxiliary_variables(fixed_parameters, profiled_parameters)
        if update_snapshot:
            self.save_snapshot(self.kCurrentSnapshotName, WSArgument.MUTABLE)
        self._check_poi_setup(fixed_parameters, profiled_parameters)
    
    def _update_floating_auxiliary_variables(self, fixed_parameters:List, profiled_parameters:List):
        if self.model.floating_auxiliary_variables is None:
            self.model._load_floating_auxiliary_variables()
        fixed_parameters    = RooArgSet.from_list(fixed_parameters)
        profiled_parameters = RooArgSet.from_list(profiled_parameters)
        core_variables = self.get_variables(WSArgument.CORE)
        fixed_parameters.remove(core_variables)
        profiled_parameters.remove(core_variables)
        self.model._floating_auxiliary_variables.add(profiled_parameters)
        self.model._floating_auxiliary_variables.remove(fixed_parameters)
        
    def _check_poi_setup(self, fixed_parameters:List, profiled_parameters:List):
        # pois defined in workspace
        ws_pois  = self.model.pois
        # pois defined in this analysis
        ana_pois = ROOT.RooArgSet(self.poi)
        if ana_pois.size() == 0:
            return None
        aux_pois = ROOT.RooArgSet(ws_pois)
        aux_pois.remove(ana_pois)
        for profiled_param in profiled_parameters:
            if aux_pois.contains(profiled_param):
                param_name = profiled_param.GetName()
                self.stdout.warning(f"WARNING: Attemping to profile the parameter \"{param_name}\" which is a "
                                    f"POI defined in the workspace but not a POI used in this study. This "
                                    f"parameter will still be fixed during likelihood fit / limit setting. "
                                    f"Please designate the parameter as a POI in this study if intended.", "red")
        for fixed_param in fixed_parameters:
            if ana_pois.contains(fixed_param):
                param_name = fixed_param.GetName()
                self.stdout.warning(f"WARNING: Attemping to fix the parameter \"{param_name}\" which is a "
                                    f"POI used in this study. This parameter will still be floated during "
                                    f"unconditional likelihood fit / limit setting.", "red")
            
    def setup_minimizer(self, constrain_nuis:bool=True, **kwargs):
        
        minimizer = self.minimizer_cls("Minimizer", self.model.pdf, self.model.data,
                                       workspace=self.model.workspace,
                                       verbosity=self.stdout.verbosity)
        
        nll_options = {k:v for k,v in kwargs.items() if k in ExtendedMinimizer._DEFAULT_NLL_OPTION_}
        
        if constrain_nuis:
            nll_options['constrain'] = self.model.nuisance_parameters
            nll_options['global_observables'] = self.model.global_observables
            conditional_obs = self.model.model_config.GetConditionalObservables()
            if conditional_obs:
                nll_options['conditional_observables'] = conditional_obs
            #nll_commands.append(ROOT.RooFit.ExternalConstraints(ROOT.RooArgSet()))
        
        minimizer_options = {k:v for k,v in kwargs.items() if k in ExtendedMinimizer._DEFAULT_MINIMIZER_OPTION_}

        minimizer.configure_nll(**nll_options)
        minimizer.configure(**minimizer_options)
        self.minimizer = minimizer
        self.default_nll_options = nll_options
        self.default_minimizer_options = minimizer_options
    
    def set_data(self, data_name:str='combData'):
        data = self.model.workspace.data(data_name)
        if not data:
            raise RuntimeError(f'workspace does not contain the dataset "{data_name}"')
        self.minimizer.set_data(data)
        self.model._data = data
        
    def set_blind_range(self, blind_range:List[float], categories:Optional[List[str]]=None):
        self.model.create_blind_range(blind_range, categories)
        sideband_range_name = self.model.get_sideband_range_name()
        self.minimizer.configure_nll(range=sideband_range_name, split_range=True, update=True)
        self._use_blind_range = True
        
    def unset_blind_range(self):
        self.minimizer.nll = None
        self.minimizer.nll_commands.pop("RangeWithName", None)
        self.minimizer.nll_commands.pop("SplitRange", None)
        self.stdout.info("INFO: Blind range removed from list of  NLL commands. NLL is reset.")
        self._use_blind_range = False
        
    def load_snapshot(self, snapshot_name:Optional[str]=None):
        if snapshot_name is not None:
            snapshot = self.model.workspace.getSnapshot(snapshot_name)
            if (not snapshot) and ('0x(nil)' in snapshot.__repr__()):
                self.stdout.warning(f'WARNING: Snapshot "{snapshot_name}" does not exist.')
            else:
                self.model.workspace.loadSnapshot(snapshot_name)
                self.stdout.debug(f'DEBUG: Loaded snapshot "{snapshot_name}"')
                
    def restore_snapshot(self):
        self.load_snapshot(self.kCurrentSnapshotName)
    
    def get_variables(self, variable_type:Union[str, WSArgument], sort:bool=True):
        return self.model.get_variables(variable_type, sort=sort)
    
    def save_snapshot(self, snapshot_name:Optional[str]=None, 
                      variables:Optional[Union[ROOT.RooArgSet, str, WSArgument]]=None):
        if snapshot_name is not None:
            if variables is None:
                self.model.workspace.saveSnapshot(snapshot_name, self.model.workspace.allVars())
            else:
                if isinstance(variables, (str, WSArgument)):
                    variables = self.get_variables(variables)
                self.model.workspace.saveSnapshot(snapshot_name, variables)
            self.stdout.debug(f'DEBUG: Saved snapshot "{snapshot_name}"')
        
    def save(self, filename:str, recreate:bool=True, rebuild:bool=True):
        self.model.save(filename, recreate=recreate, rebuild=rebuild)
        self.stdout.info(f'INFO: Saved workspace file as "{filename}"')
        
        
    def decompose_nll(self, fmt:str="pandas"):
        from quickstats.utils.roofit_utils import decompose_nll
        if not self.minimizer.nll:
            self.minimizer.create_nll()
            result = decompose_nll(self.minimizer.nll, self.model.global_observables, fmt=fmt)
            self.minimizer.nll = None
            return result
        return decompose_nll(self.minimizer.nll, self.model.global_observables, fmt=fmt)


from typing import Optional, List, Dict
import os
import re

class RooProcBaseAction(object):
    def __init__(self, **params):
        self._params = params
        self.executed = False
        self.status   = None
    
    def get_formatted_parameters(self, global_vars:Optional[Dict]=None):
        if global_vars is None:
            global_vars = {}
        formatted_parameters = {}
        for k,v in self._params.items():
            if v is None:
                formatted_parameters[k] = None
                continue
            k_literals = re.findall(r"\${(\w+)}", k)
            is_list = False
            if isinstance(v, list):
                v = '__SEPARATOR__'.join(v)
                is_list = True
            v_literals = re.findall(r"\${(\w+)}", v)
            all_literals = set(k_literals).union(set(v_literals))
            for literal in all_literals:
                if literal not in global_vars:
                    raise RuntimeError(f"the global variable `{literal}` is undefined")
            for literal in k_literals:
                substitute = global_vars[literal]
                k = k.replace("${" + literal + "}", str(substitute))
            for literal in v_literals:
                substitute = global_vars[literal]
                v = v.replace("${" + literal + "}", str(substitute))
            if is_list:
                v = v.split("__SEPARATOR__")
            formatted_parameters[k] = v
        return formatted_parameters
    
    def makedirs(self, fname:str):
        dirname = os.path.dirname(fname)
        if dirname and (not os.path.exists(dirname)):
            os.makedirs(dirname)
    
    def execute(self, **params):
        raise NotImplementedError
    
    @classmethod
    def parse_as_list(cls, text:str):
        match = re.match(r"\[([^\[\]]+)\]", text)
        if not match:
            return [text]
        else:
            return match.group(1).split(",")
    
    @classmethod
    def parse_as_kwargs(cls, text:str):
        kwargs = {}
        text = re.sub(r"\s*", "", text)
        list_attributes = re.findall(r"(\w+)=\[([^\[\]]+)\]", text)
        for attribute in list_attributes:
            kwargs[attribute[0]] = attribute[1].split(",")
            text = text.replace(f"{attribute[0]}=[{attribute[1]}]","")
        attributes = re.findall(r"(\w+)=([^,]+)", text)
        for attribute in attributes:
            kwargs[attribute[0]] = attribute[1]
        return kwargs
    
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        raise NotImplementedError


from typing import Optional, Union, Dict, List, Any

from quickstats.utils.common_utils import combine_dict
from quickstats.plots import AbstractPlot

class CollectiveDataPlot(AbstractPlot):
    
    def __init__(self, collective_data:Dict[str, Any],
                 plot_options:Optional[Dict[str, Dict]]=None,
                 label_map:Optional[Dict[str, str]]=None,
                 color_pallete:Optional[Dict]=None,
                 color_cycle:Optional[Dict]=None,
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Union[Dict, str]]=None,
                 figure_index:Optional[int]=None,
                 config:Optional[Dict]=None):
        
        super().__init__(color_pallete=color_pallete,
                         color_cycle=color_cycle,
                         styles=styles, 
                         analysis_label_options=analysis_label_options,
                         figure_index=figure_index,
                         config=config)
        
        self.set_data(collective_data)
        self.plot_options = combine_dict({}, plot_options)
        self.label_map    = combine_dict({}, label_map)
        
    def set_data(self, collective_data:Dict[str, Any]):
        if not isinstance(collective_data, dict):
            raise ValueError("data collection must be contained in a dictionary")
        resolved_data = {}
        for label, data in collective_data.items():
            resolved_data[label] = self.parse_data(data)
        self.collective_data = resolved_data
    
    @classmethod
    def parse_data(cls, data:Any):
        return data


from typing import Optional, Union, Dict, List, Tuple, Callable

from cycler import cycler

import numpy as np
import pandas as pd

from quickstast import semistaticmethod
from quickstats.plots.color_schemes import QUICKSTATS_PALETTES

from quickstats.plots import AbstractPlot, CollectiveDataPlot
from quickstats.plots.template import suggest_markersize, ratio_frames, centralize_axis, create_transform
from quickstats.utils.common_utils import combine_dict
from quickstats import GeneralEnum

class ErrorDisplayFormat(GeneralEnum):
    ERRORBAR = 0
    FILL     = 1

class GeneralDistributionPlot(AbstractPlot):
    
    STYLES = {
        'errorbar': {
            "marker": 'o',
            "markersize": None,
            "linewidth": 0,
            "elinewidth": 1,
            "capsize": 3,
            "capthick": 1
        },
        "annotation": {
            "xy": (0.05, 0.9),
            "xycoords": "axes fraction",
            "horizontalalignment": "left",
            "verticalalignment": "top",
            "fontsize": 20
        },
        "legend": {
            "borderpad": 1
        },
        "ratio_frames": {
            "height_ratios": (4, 1)
        }
    }
    
    CONFIG = {
        "blind_region_styles":{
            "linestyle": "--"
        },
        'ratio_line_styles':{
            'color': 'gray',
            'linestyle': '--'
        }        
    }
    
    def __init__(self, collective_data:Dict[str, Union[Dict[str, np.ndarray], pd.DataFrame]],
                 blind_ranges:Optional[Union[List[float], List[List[float]]]],
                 plot_options:Optional[Dict[str, Dict]]=None,
                 scale_map:Optional[Dict[str, float]]=None,
                 label_map:Optional[Dict[str, str]]=None,
                 color_cycle:Optional[Dict]=None,
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Union[Dict, str]]=None,
                 figure_index:Optional[int]=None,
                 config:Optional[Dict]=None):
        """
        
        Parameters:
            collective_data: dictionary
                Input data of the form {data_label: {"x": array, "y": array}, ...}.
                To specify symmetric errors, use {"x": array, "y": array, "xerr": array, "yerr": array}.
                To specify asymmetric errors, use {"x": array, "y": array,
                "xerrlo": array, "xerrhi": array, "yerrlo": array, "yerrhi": array}.
        """

        self.error_options = {}
        self.scale_map     = combine_dict({}, scale_map)
        self.colors        = {} 
        self.annotation    = None
        
        self.blind_ranges(blind_ranges)
        
        super().__init__(collective_data=collective_data,
                         plot_options=plot_options,
                         label_map=label_map,
                         color_cycle=color_cycle,
                         styles=styles, 
                         analysis_label_options=analysis_label_options,
                         figure_index=figure_index,
                         config=config)
        
    def set_blind_ranges(self, blind_ranges:Optional[Union[List[float], List[List[float]]]]=None):
        if blind_ranges is None:
            self.blind_ranges = None
            return None
        from quickstats.maths.numerics import get_proper_ranges
        self.blind_ranges = get_proper_ranges(blind_ranges)
        
    def set_data(self, collective_data:Dict[str, "pandas.DataFrame"]):
        super().set_data(collective_data)
        self.legend_order = list(collective_data.keys())
        for name in self.collective_data:
            if name not in self.error_options:
                self.set_error_options(name)        
        
    @classmethod
    def parse_data(cls, data:Dict[str, Union[Dict[str, np.ndarray], pd.DataFrame]]):
        if not isinstance(data, (dict, pd.DataFrame)):
            raise ValueError("input data must be a dataframe or dictionary")
        for attribute in ["x", "y"]:
            if attribute not in data:
                raise RuntimeError(f'missing attribute "{attribute}" from input data')
        x, y = np.array(data['x']), np.array(data['y'])
        error_data = cls.get_error_data(data)
        # check shape consistency
        if x.shape != y.shape:
            raise ValueError('invalid input: "x" and "y" values must have the same shape')
        data_shape = x.shape
        for label in ["xerr", "yerr"]:
            err_values = error_data[label]
            if err_values is None:
                continue
            if isinstance(err_values, tuple):
                valid_shape = (error_values[0].shape == data_shape) and (error_values[1].shape == data_shape)
            else:
                valid_shape = error_values.shape == data_shape
            if not valid_shape:
                raise RuntimeError('inconsistent shape between data and error values')
        parsed_data = {
            'x'    : x,
            'y'    : y,
            'xerr' : error_data['xerr']
            'yerr' : error_data['yerr']
        }
        return parsed_data           

    @semistaticmethod
    def get_error_data(data:Union[Dict[str, np.ndarray], pd.DataFrame]):
        error_data = {}
        for label in ["xerr", "yerr"]:
            err   = data.get(label, None)
            errlo = data.get(f"{label}lo", None)
            errhi = data.get(f"{label}hi", None)
            # set default to None
            error_data[label] = None
            if (err is None):
                if  (errlo is not None) and (errhi is None):
                    values = np.array(errlo)
                    error_data[label] = (values, np.zeros(values.shape))
                elif (errlo is None) and (errhi is not None):
                    values = np.array(errhi)
                    error_data[label] = (np.zeros(values.shape), values)
                elif (errlo is not None) and (errhi is not None):
                    error_data[label] = (np.array(errlo), np.array(errhi))
            else:
                if ((errlo is not None) or (errhi is not None)):
                    raise ValueError(f'invalid input: received both "{label}" and '
                                     f'"{label}lo"/"{label}hi" attributes')
                error_data[label] = np.array(err)
        return error_data
    
    @semistaticmethod
    def select_data(self, data:Dict[str, np.ndarray], key:str,
                    value_ranges:np.ndarray):
        if err is None:
            return None
        elif isinstance(err, tuple):
            new_err = []
            for err_i in err:
                new_err.append(err_i[mask])
            return tuple(new_err)
        else:
            return err[mask]

    def draw_single_data(self, ax, data, label:str, label_blind:Optional[str]=None,
                         blind_range:Optional[List[float]]=None,
                         plot_options:Optional[Dict]=None, show_error:bool=True,
                         error_format:Union[ErrorDisplayFormat, str]="error_bar",
                         error_styles:Optional[Dict]=None,
                         scale_by:Optional[float]=None):
        pdata = self.process_data(data, blind_range=blind_range, scale_by=scale_by)
        has_errorbar = not ((pdata['xerr'] is None) and (pdata['yerr'] is None))
        draw_blind = blind_range is not None
        if not has_errorbar:
            combined_options = combine_dict(self.styles['plot'], plot_options)
            combined_options['label'] = label
            if draw_blind:
                # draw sideband low
                handle_sblo = ax.plot(pdata['x'][pdata['sideband_lo']], 
                                      pdata['y'][pdata['sideband_lo']],
                                      **combined_options)
                # avoid going through internal color cycle for the three regions
                combined_options['color'] = handle_sblo[0].get_color()
                # draw sideband high
                handle_sbhi = ax.plot(pdata['x'][pdata['sideband_hi']], 
                                      pdata['y'][pdata['sideband_hi']],
                                      **combined_options)
                combined_options['linestyle'] = self.config['blind_region_styles']['linestyle']
                if label_blind is None:
                    combined_options['label'] = f"{label} (blind)"
                else:
                    combined_options['label'] = label_blind
                x_blind_plus_sideband_edge = [pdata['x'][pdata['sideband_lo']][-1]] + \
                                             list(pdata['x'][pdata['blind']]) + \
                                             [pdata['x'][pdata['sideband_hi']][0]]
                y_blind_plus_sideband_edge = [pdata['y'][pdata['sideband_lo']][-1]] + \
                                             list(pdata['y'][pdata['blind']]) + \
                                             [pdata['y'][pdata['sideband_hi']][0]]                
                handle_blind = ax.plot(x_blind_plus_sideband_edge, 
                                       y_blind_plus_sideband_edge,
                                       **combined_options)
                handle = [handle_sblo[0], handle_blind[0]]
            else:
                handle = ax.plot(pdata['x'], pdata['y'], **combined_options)
                handle = handle[0]
        else:
            
            combined_options = combine_dict(plot_options, {"label": label})
            if combined_options.get('markersize', None) is None:
                nbins = len(pdata['x'])
                combined_options['markersize'] = suggest_markersize(nbins)
                
            if draw_blind:
                x = pdata['x'][pdata['sideband']]
                y = pdata['y'][pdata['sideband']]
                xerr = self._get_selected_err(pdata['xerr'], pdata['sideband'])
                yerr = self._get_selected_err(pdata['yerr'], pdata['sideband'])
            else:
                x = pdata['x']
                y = pdata['y']
                xerr = pdata['xerr']
                yerr = pdata['yerr']
            if not show_error:
                error_format = None
            handle = self._draw_data(ax, x, y, xerr, yerr,
                                     plot_options=combined_options,
                                     error_format=error_format,
                                     error_styles=error_styles)
        return handle
    
    def _draw_data(self, ax, x:np.ndarray, y:np.ndarray,
                   xerr:Optional[np.ndarray]=None,
                   yerr:Optional[np.ndarray]=None,
                   plot_options:Optional[Dict]=None,
                   error_format:Optional[Union[ErrorDisplayFormat, str]]="error_bar",
                   error_styles:Optional[Dict]=None):
        combined_options = combine_dict(plot_options, self.styles['errorbar'])
        if error_format is None:
            combined_options['elinewidth'] = 0
            combined_options['capsize'] = 0
            combined_options['capthick'] = 0
            handle = ax.errorbar(x, y, xerr=xerr, yerr=yerr,
                                 **combined_options)
        elif (error_format == ErrorDisplayFormat.ERRORBAR):
            combined_options = combine_dict(combined_options, error_styles)
            handle = ax.errorbar(x, y, xerr=xerr, yerr=yerr,
                                 **combined_options)
        elif (error_format == ErrorDisplayFormat.FILL):
            combined_options['elinewidth'] = 0
            combined_options['capsize'] = 0
            combined_options['capthick'] = 0
            handle = ax.errorbar(x, y, xerr=xerr, yerr=yerr,
                                 **combined_options)
            fill_options = combine_dict(self.styles['fill_between'], error_styles)
            if "zorder" not in fill_options:
                fill_options["zorder"] = 5
            if yerr is not None:
                handle_fill = ax.fill_between(x, y - yerr[0], y + yerr[1],
                                              **fill_options)
            else:
                handle_fill = None
        else:
            raise RuntimeError(f"unkown error display format: {error_format}")
        return handle
    
    def draw_comparison(self, ax, reference_data, target_data, 
                        xlabel:str="", ylabel:str="",  mode:str="ratio",
                        plot_options:Optional[Dict]=None, show_error:bool=True, 
                        blind_range:Optional[List[float]]=None,
                        error_format:Union[ErrorDisplayFormat, str]="error_bar",
                        error_styles:Optional[Dict]=None,                        
                        draw_ratio_line:bool=True):
        pdata_ref = self.process_data(reference_data, blind_range=blind_range)
        pdata_tgt = self.process_data(target_data, blind_range=blind_range)
        if not np.allclose(pdata_tgt['x'], pdata_ref['x']):
            raise RuntimeError("cannot compare two distributions with different binnings")
        pdata_comp = {}
        pdata_comp['x'] = pdata_ref['x']
        pdata_comp['yerr'] = 0
        pdata_comp['xerr'] = pdata_ref['xerr']
        if pdata_ref['yerr'] is None:
            if isinstance(pdata_tgt['yerr'], tuple):
                yerr_ref = (np.zeros(pdata_ref['y'].shape), np.zeros(pdata_ref['y'].shape))
            else:
                yerr_ref = np.zeros(pdata_ref['y'].shape)
        else:
             yerr_ref = pdata_ref['yerr']
        if pdata_tgt['yerr'] is None:
            if isinstance(pdata_ref['yerr'], tuple):
                yerr_tgt = (np.zeros(pdata_tgt['y'].shape), np.zeros(pdata_tgt['y'].shape))
            else:
                yerr_tgt = np.zeros(pdata_tgt['y'].shape)
        else:
             yerr_tgt = pdata_tgt['yerr']
        if mode == "ratio":
            vmode = 0
        elif mode in ["diff", "difference"]:
            vmode = 1
        else:
            raise ValueError(f"unsupported mode \"{mode}\", choose between \"ratio\" and \"difference\"")
        if vmode == 0:
            pdata_comp['y'] = pdata_tgt['y'] / pdata_ref['y']
            if isinstance(yerr_ref, tuple):
                errlo = np.sqrt((yerr_ref[0]/pdata_ref['y'])**2 + (yerr_tgt[0]/pdata_tgt['y'])**2)
                errhi = np.sqrt((yerr_ref[1]/pdata_ref['y'])**2 + (yerr_tgt[1]/pdata_tgt['y'])**2)
                pdata_comp['yerr'] = (errlo, errhi)
            else:
                pdata_comp['yerr'] = np.sqrt((yerr_ref/pdata_ref['y'])**2 + (yerr_tgt/pdata_tgt['y'])**2)
        elif vmode == 1:
            pdata_comp['y'] = pdata_tgt['y'] - pdata_ref['y']
            if isinstance(yerr_ref, tuple):
                errlo = np.sqrt(yerr_ref[0]**2 + yerr_tgt[0]**2)
                errhi = np.sqrt(yerr_ref[1]**2 + yerr_tgt[1]**2)
                pdata_comp['yerr'] = (errlo, errhi)
            else:
                pdata_comp['yerr'] = np.sqrt(yerr_ref**2 + yerr_tgt**2)
            
        combined_options = combine_dict(plot_options)
        if combined_options.get('markersize', None) is None:
            nbins = len(pdata_comp['x'])
            combined_options['markersize'] = suggest_markersize(nbins)
            
        draw_blind = blind_range is not None
        if draw_blind:
            x = pdata_comp['x'][pdata_ref['sideband']]
            y = pdata_comp['y'][pdata_ref['sideband']]
            xerr = self._get_selected_err(pdata_comp['xerr'], pdata_ref['sideband'])
            yerr = self._get_selected_err(pdata_comp['yerr'], pdata_ref['sideband'])
        else:
            x = pdata_comp['x']
            y = pdata_comp['y']
            xerr = pdata_comp['xerr']
            yerr = pdata_comp['yerr']
            
        if not show_error:
            error_format = None
        handle = self._draw_data(ax, x, y, xerr, yerr,
                                 plot_options=combined_options,
                                 error_format=error_format,
                                 error_styles=error_styles)
            
        ylim = list(ax.get_ylim())
        if ylim[0] > np.min(y):
            ylim[0] = np.min(y)
        if ylim[1] < np.max(y):
            ylim[1] = np.max(y)
        ax.set_ylim(ylim)
        
        if vmode == 0:
            centralize_axis(ax, which="y", ref_value=1)
            if draw_ratio_line:
                ratio_linestyles = self.config['ratio_line_styles']
                ax.axhline(1, zorder=0, **ratio_linestyles)
        elif vmode == 1:
            centralize_axis(ax, which="y", ref_value=0)
            if draw_ratio_line:
                ratio_linestyles = self.config['ratio_line_styles']
                ax.axhline(0, zorder=0, **ratio_linestyles)
        if not ylabel:
            if vmode == 0:
                ylabel = "Fit / Data"
            elif vmode == 1:
                ylabel = "Fit - Data"
            else:
                ylabel = ""
        self.draw_axis_components(ax, xlabel=xlabel, ylabel=ylabel)
        return handle
        
    def set_error_options(self, target:str,
                          display_format:Optional[Union[ErrorDisplayFormat, str]]="errorbar",
                          styles:Optional[Dict]=None):
        if display_format is not None:
            display_format = ErrorDisplayFormat.parse(display_format)
        self.error_options[target] = {
            "display_format": display_format,
            "styles": styles
        }

    def draw(self, xlabel:str="", ylabel:str="",
             targets:Optional[List]=None,
             logx:bool=False, logy:bool=False,
             blind_range:Optional[List[float]]=None,
             show_error:bool=True,
             comparison_options:Optional[Union[Dict, List[Dict]]]=None,
             xmin:Optional[float]=None, xmax:Optional[float]=None,
             ymin:Optional[float]=None, ymax:Optional[float]=None,
             ypad:Optional[float]=None):
        
        if comparison_options is not None:
            ax, ax_ratio = self.draw_frame(ratio_frames, logx=logx, logy=logy,
                                           **self.styles["ratio_frames"])
        else:
            ax = self.draw_frame(logx=logx, logy=logy)

        for name in self.collective_data:
            if (targets is not None) and(name not in targets):
                continue
            data = self.collective_data[name]
            label = self.label_map.get(name, name)
            label_blind = self.label_map.get(f"{name}_blind", None)
            plot_options = self.plot_options.get(name, {})
            scale_by = self.scale_map.get(name, None)
            if (name not in self.error_options) or \
               (self.error_options[name]["display_format"] is None):
                error_format = None
                error_styles = None
            else:
                error_format = self.error_options[name]["display_format"]
                error_styles = self.error_options[name]["styles"]
            handle = self.draw_single_data(ax, data, label,
                                           label_blind=label_blind,
                                           blind_range=blind_range,
                                           plot_options=plot_options,
                                           show_error=show_error,
                                           error_format=error_format,
                                           error_styles=error_styles,
                                           scale_by=scale_by)
            # case draw blind
            if isinstance(handle, list) and len(handle) == 2:
                blind_name = f"{name}_blind"
                self.legend_order.append(blind_name)
                self.update_legend_handles({name: handle[0], blind_name: handle[1]})
                self.colors[name] = handle[0].get_color()
            else:
                self.update_legend_handles({name: handle})
                if not handle.get_children():
                    self.colors[name] = handle.get_color()
                else:
                    self.colors[name] = handle[0].get_color()
        
        self.draw_axis_components(ax, xlabel=xlabel, ylabel=ylabel)
        self.set_axis_range(ax, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, ypad=ypad)  
        handles, labels = self.get_legend_handles_labels()
        ax.legend(handles, labels, **self.styles['legend'])
        
        if comparison_options is not None:
            if not isinstance(comparison_options, list):
                comparison_options = [comparison_options]
            for options in comparison_options:
                reference = options['reference']
                target    = options['target']
                mode      = options.get('mode', "ratio")
                ylabel    = options.get('label', "")
                labelsize = options.get('labelsize', None)
                styles    = options.get('styles', {})
                draw_ratio_line = options.get('draw_ratio_line', True)
                plot_options = self.plot_options.get(reference, {})
                if 'color' not in plot_options:
                    plot_options['color'] = self.colors[target]
                plot_options.update(styles)
                if (reference not in self.error_options) or \
                   (self.error_options[reference]["display_format"] is None):
                    error_format = None
                    error_styles = None
                else:
                    error_format = self.error_options[reference]["display_format"]
                    error_styles = self.error_options[reference]["styles"]
                self.draw_comparison(ax_ratio,
                                     self.collective_data[reference],
                                     self.collective_data[target],
                                     xlabel=ax.get_xlabel(),
                                     ylabel=ylabel, mode=mode,
                                     plot_options=plot_options,
                                     show_error=show_error,
                                     error_format=error_format,
                                     error_styles=error_styles,
                                     blind_range=blind_range,
                                     draw_ratio_line=draw_ratio_line)
                if labelsize is not None:
                    ax_ratio.yaxis.label.set_size(labelsize)                
            ax.set(xlabel=None)
            ax.tick_params(axis="x", labelbottom=False)

        if self.annotation:
            kwargs = combine_dict(self.styles['annotation'], self.annotation)
            ax.annotate(**kwargs)
            
        if comparison_options is not None:
            return ax, ax_ratio
        
        return ax



from .decorators import *
from .methods import *
from .abstract_object import AbstractObject
from .enums import GeneralEnum, DescriptiveEnum
from .virtual_trees import TVirtualNode, TVirtualTree
from .path_manager import DynamicFilePath, PathManager
from .configs import ConfigComponent, ConfigParser
from .configurable_object import ConfigurableObject
from .dataclass_object import DataclassObject


###############################################################################
### This is a reimplementation of xmlAnaWSBuilder library in python
### original author: Hongtao Yang
###############################################################################
import os
import re
import time
from typing import Optional, Union, List, Dict

import numpy as np

import ROOT

from quickstats import semistaticmethod, DescriptiveEnum, PathManager
from quickstats.components import ExtendedModel
from quickstats.utils.io import Verbosity
from quickstats.utils.xml_tools import TXMLTree
from quickstats.utils.common_utils import remove_list_duplicates, format_delimiter_enclosed_text
from quickstats.maths.numerics import is_float, pretty_value
from quickstats.interface.root import TH1, RooDataSet

from quickstats.components.workspaces import (XMLWSBase, Sample, Systematic, SystematicType,
                                              SystematicsDomain, AsimovHandler, CoreArgumentSets)

from .settings import *

class DataStorageType(DescriptiveEnum):
    Tree      = (0, "Store data as a tree")
    Vector    = (1, "Store data as a vector")
    Composite = (2, "Store data as a mixture of tree and vector")

class SampleFactorType(DescriptiveEnum):
    NormFactor  = (0, "Normalization factor")
    ShapeFactor = (1, "Shape factor")
    
class XMLWSBuilder(XMLWSBase):
    
    EPSILON = 1e-6
    
    def __init__(self, source:str, basedir:Optional[str]=None,
                 use_binned:bool=False,
                 data_storage_type:str="vector",
                 minimizer_config:Optional[Dict]=None,
                 unlimited_stack:bool=True,
                 verbosity:Optional[Union[int, str]]="INFO") -> None:
        super().__init__(source=source, basedir=basedir,
                         unlimited_stack=unlimited_stack,
                         verbosity=verbosity)
        self.use_binned = use_binned
        self.data_storage_type = data_storage_type
        self.minimizer_config = minimizer_config
        self.load_extension()
        self.initialize(source=source)

    def initialize(self, source:str) -> None:
        self.core_config = self.parse_core_xml(source)
        self.update_paths()
        self.reset()
        self.set_data_storage_type(self.data_storage_type)
        self.set_integrator(self.core_config['integrator'])
        self.print_init_summary()
        
    def reset(self) -> None:
        self.cat_config      = {}
        self.cat_ws          = {}
        self.datahists       = {}
        self.datasets        = {}
        self.datasets_binned = {}
        self.time_taken      = 0.
        ws_name = self.core_config['workspace_name']
        mc_name = self.core_config['model_config_name']
        self.comb_ws      = ROOT.RooWorkspace(ws_name)
        self.model_config = ROOT.RooStats.ModelConfig(mc_name, self.comb_ws)
        
    def update_paths(self) -> None:
        self.path_manager.set_file("output", self.core_config["output_name"])
        categories = []
        for category, file in self.core_config['input_categories'].items():
            self.path_manager.set_file(f"__Cat_{category}", file)
                                       
    @semistaticmethod
    def set_data_storage_type(self, data_storage_type:Union[str, DataStorageType]="vector") -> None:
        data_storage_type = DataStorageType.parse(data_storage_type)
        self.stdout.info(f'INFO: Set data storage type to "{data_storage_type.name}"')
        if data_storage_type == DataStorageType.Tree:
            ROOT.RooAbsData.setDefaultStorageType(ROOT.RooAbsData.StorageType.Tree)
        elif data_storage_type == DataStorageType.Vector:
            ROOT.RooAbsData.setDefaultStorageType(ROOT.RooAbsData.StorageType.Vector)
        elif data_storage_type == DataStorageType.Composite:
            ROOT.RooAbsData.setDefaultStorageType(ROOT.RooAbsData.StorageType.Composite)            
        else:
            raise RuntimeError(f'unknown data storage type "{data_storage_type}"')
            
    @semistaticmethod
    def set_integrator(self, integrator:Optional[str]=None) -> None:
        if integrator:
            self.stdout.warn(f'WARNING: Set default integrator to "{integrator}"')
            ROOT.RooAbsReal.defaultIntegratorConfig().method1D().setLabel(integrator)
    
    def parse_core_xml(self, filename:str) -> Dict:
        self.stdout.info(f'INFO: Parsing file "{filename}"')
        # load xml file
        core_xml = TXMLTree.load_as_dict(filename)
        # parse core attributes
        attributes = core_xml['attrib']
        core_config = {
            'base_dir'           : os.path.dirname(filename),
            'workspace_name'     : self._get_node_attrib(core_xml, 'WorkspaceName'),
            'model_config_name'  : self._get_node_attrib(core_xml, 'ModelConfigName'),
            'dataset_name'       : self._get_node_attrib(core_xml, 'DataName'),
            'output_name'        : self._get_node_attrib(core_xml, 'OutputFile'),
            'do_blind'           : self._get_node_attrib(core_xml, 'Blind',
                                                         required=False, default=False, dtype="bool"),
            'scale_lumi'         : self._get_node_attrib(core_xml, 'ScaleLumi',
                                                         required=False, default=-1., dtype="float"),
            'integrator'         : self._get_node_attrib(core_xml, 'Integrator',
                                                         required=False, default=""),
            'generate_hist_data' : self._get_node_attrib(core_xml, 'GenHistData',
                                                         required=False, default=False, dtype="bool"),
            'poi_names'          : [],
            'input_categories'   : {},
            'asimov_definitions' : []
        }
        # config for import class codes
        class_decl_import_dir = self._get_node_attrib(core_xml, 'ClassDeclImportDir',
                                                      required=False, default='inc/')
        class_decl_import_dir = self.path_manager.get_relpath(class_decl_import_dir)
        class_impl_import_dir = self._get_node_attrib(core_xml, 'ClassImplImportDir',
                                                      required=False, default='src/')
        class_impl_import_dir = self.path_manager.get_relpath(class_impl_import_dir)
        if os.path.isdir(class_decl_import_dir) and os.path.isdir(class_impl_import_dir):
            core_config['class_code_dirs'] = {
                'decl': class_decl_import_dir,
                'impl': class_impl_import_dir,
            }
        elif (not os.path.isdir(class_decl_import_dir)) and (not os.path.isdir(class_impl_import_dir)):
            core_config['class_code_dirs'] = {
                'decl': None,
                'impl': None,
            }
        else:
            raise FileNotFoundError("directories for importing class code do not exist "
                                    f"(decl = `{class_decl_import_dir}`, impl = `{class_impl_import_dir}`)")
        # make sure output filename has .root extension
        if not core_config['output_name'].endswith(".root"):
            core_config['output_name'] = os.path.splitext(core_config['output_name'])[0] + ".root"
            self.stdout.warning('WARNING: Output file name does not contain ".root" postfix. '
                                'Adding it to avoid confusion.')
        # parse child level nodes
        nodes = core_xml['children']
        for node in nodes:
            tag = node['tag']
            if tag == "Input":
                file = node['text']
                tree = TXMLTree(file=self.path_manager.get_relpath(file))
                category = tree.root.get_attribute("Name")
                if category in core_config['input_categories']:
                    raise RuntimeError(f'duplicated category name: "{category}"')
                core_config['input_categories'][category] = file
            elif tag == "POI":
                poi_expr = node['text']
                core_config['poi_names'] += poi_expr.split(",")
            elif tag == "Asimov":
                definitions = node['attrib']
                core_config['asimov_definitions'].append(definitions)
            else:
                raise RuntimeError(f"unknown item `{tag}`")
        return core_config
    
    def print_init_summary(self) -> None:
        if not self.core_config:
            raise RuntimeError("core config not set")
        self.stdout.info("="*74)
        self.stdout.info("Output file: ".rjust(20) + f"{self.core_config['output_name']}")
        self.stdout.info("Workspace name: ".rjust(20) + f"{self.core_config['workspace_name']}")
        self.stdout.info("ModelConfig name: ".rjust(20) + f"{self.core_config['model_config_name']}")
        self.stdout.info("Dataset name: ".rjust(20) + f"{self.core_config['dataset_name']}")
        self.stdout.info("Blind analysis: ".rjust(20) + ("True" if self.core_config['do_blind'] else "False"))
        self.stdout.info("Generate hist data: ".rjust(20) + ("True" if self.core_config['generate_hist_data'] else "False"))
        self.stdout.info("Scale lumi: ".rjust(20) + f"{self.core_config['scale_lumi']}")
        self.stdout.info("POIs: ".rjust(20) + ", ".join(self.core_config['poi_names']))
        self.stdout.info("")
        self.stdout.info("Categories to be included:")
        input_categories = self.core_config['input_categories']
        for i,input_category in enumerate(input_categories.values()):
            self.stdout.info(f"XML file {i}: ".rjust(20) + f"{input_category}")
        asimov_definitions = self.core_config['asimov_definitions']
        if asimov_definitions:
            self.stdout.info("")
            self.stdout.info("The following asimov dataset(s) will be generated:")
            for asimov_def in asimov_definitions:
                is_gen_asimov = GEN_ASIMOV_ACTION in asimov_def['Action']
                if is_gen_asimov:
                    asimov_name = asimov_def['Name']
                    self.stdout.info(f"\t{asimov_name}")
        self.stdout.info("="*74)

    def print_final_summary(self) -> None:
        ws_name = self.comb_ws.GetName()
        model = ExtendedModel(self.comb_ws, data_name=self.core_config['dataset_name'], verbosity="WARNING")
        pdf = model.pdf
        cat = pdf.indexCat()
        n_cat = len(cat)
        data = model.data
        data_list = data.split(cat, True)
        # start writing summary
        self.stdout.info("="*74)
        title_str = format_delimiter_enclosed_text("Begin summary", "~")
        self.stdout.info(title_str)
        self.stdout.info(f"\tThere are {n_cat} sub channels:")
        for i in range(n_cat):
            cat.setBin(i)
            cat_name = cat.getLabel()
            pdf_i = pdf.getPdf(cat_name)
            data_i = data_list.FindObject(cat_name)
            sum_entries = pretty_value(data_i.sumEntries(), 5)
            self.stdout.info(f"\t\tIndex: {i}, Pdf: {pdf_i.GetName()}, Data: {data_i.GetName()}, SumEntries: {sum_entries}")
        title_str = format_delimiter_enclosed_text("POI", "-")
        self.stdout.info(title_str)
        if Verbosity.INFO >= self.stdout.verbosity:
            model.pois.Print("v")
        title_str = format_delimiter_enclosed_text("Dataset", "-")
        self.stdout.info(title_str)
        if Verbosity.INFO >= self.stdout.verbosity:
            for dataset in model.workspace.allData():
                dataset.Print()
        title_str = format_delimiter_enclosed_text("End Summary", "~")
        self.stdout.info(title_str)
        output_path = self.path_manager.get_file("output")
        self.stdout.info(f"Workspace `{ws_name}` has been successfully generated and saved in file `{output_path}`")
        #output_plot_path = self.get_output_plot_path()
        #self.stdout.info(f"Plots for each category are summarized in `{output_plot_path}`")
        self.stdout.info(f"Total time taken: {self.time_taken:.3f}s")
        self.stdout.info("="*74)
                         
    def generate_asimov(self, ws:Optional["ROOT.RooWorkspace"]=None) -> None:
        if ws is None:
            ws = self.comb_ws
        asimov_definitions = self.core_config['asimov_definitions']
        if self.use_binned:
            data_name = self.core_config['dataset_name'] + "binned"
            self.stdout.warning("REGTEST: Fitting binned dataset.", "red")
        else:
            data_name = self.core_config['dataset_name']
        range_name = self.get_fit_range_name()
        self._generate_asimov(ws, asimov_definitions=asimov_definitions,
                              data_name=data_name, range_name=range_name,
                              minimizer_config=self.minimizer_config)
        
    def get_fit_range_name(self) -> Optional[str]:
        """Get range name used in fitting. Returns None if the analysis is unblinded.
        """
        do_blind = self.core_config['do_blind']
        if do_blind:
            return self.get_sideband_range_name()
        else:
            return None
        
    def get_sideband_range_name(self) -> str:
        """Get sideband range name to be used in fitting.
        """
        range_name_SBLo = RANGE_NAME_SB_LO
        range_name_SBHi = RANGE_NAME_SB_HI
        return f"{range_name_SBLo},{range_name_SBHi}"
    
    def get_blind_range_name(self) -> str:
        """Get blind range name that are excluded from the fit.
        """
        return RANGE_NAME_BLIND
        
    def generate_single_category(self, category:str) -> "ROOT.RooWorkspace":
        """
        Generate workspace for a single category.
        
        Parameters
        ----------
        category: str
            Name of the category.
        """
        ws         = ROOT.RooWorkspace(f"__Cat_{category}")
        ws_factory = ROOT.RooWorkspace(f"factory_{category}")
        resp_func_map  = {}
        # load category xml file
        filepath = self.path_manager.get_file(f"__Cat_{category}", check_exist=True)
        cat_xml = TXMLTree.load_as_dict(filepath)
            
        title_str = format_delimiter_enclosed_text(f"Category {category}", "-")
        self.stdout.info(title_str)
        
        self.active_category = category
        category_type = self._get_node_attrib(cat_xml, 'Type').lower()
    
        # define luminosity
        if self.core_config['scale_lumi'] > 0:
            lumi_default = 1
        else:
            lumi_default = -1
        luminosity = self._get_node_attrib(cat_xml, "Lumi", required=False,
                                           default=lumi_default, dtype="float")
        if luminosity > 0:
            if self.core_config['scale_lumi'] > 0:
                lumi_expr = f"{LUMI_NAME}[{luminosity}]"
            else:
                lumi_expr = f"{LUMI_NAME}[1]"
            self.import_expression(ws_factory, lumi_expr)
        
        self.cat_config[category] = {
            'input_file': filepath,
            'category_type': category_type,
            'luminosity': luminosity,
            'observable': {'name': ""},
            'samples': {},
            'systematics': {},
            'items_highpriority': [],
            'items_lowpriority': [],
            'weighted_data': False,
            'resp_func_map': {}
        }
        core_argsets = CoreArgumentSets(verbosity=self.stdout.verbosity)
        self.cat_config[category]['core_argsets'] = core_argsets

        nodes = cat_xml['children']
        
        # define dataset
        data_node = self.fetch_node(nodes, "Data", xml_filename=filepath, xml_type="channel",
                                    allow_multiple=False, allow_missing=False)
        self.read_data_node(data_node, ws_factory)
        
        # define correlation terms        
        correlate_node = self.fetch_node(nodes, "Correlate", xml_filename=filepath, xml_type="channel",
                                         allow_multiple=False, allow_missing=True)
        self.read_correlate_node(correlate_node)

        remained_nodes = [node for node in nodes if node['tag'] not in ["Data", "Correlate"]]
        self.read_channel_xml_nodes(remained_nodes, ws_factory)
 
        self._fix_item_priority()
        items_highpriority  = self.cat_config[category]['items_highpriority']
        items_lowpriority   = self.cat_config[category]['items_lowpriority']
        
        self.stdout.debug("REGTEST: High priority items:")
        for item in items_highpriority:
            self.stdout.debug(f"\t{item}")
        self.stdout.debug("REGTEST: Low priority items:")
        for item in items_lowpriority:
            self.stdout.debug(f"\t{item}")
        # first implement high priority items: common variables 
        # and functions not involving systematic uncertainties
        self.import_expressions(ws_factory, items_highpriority)
        
        # secondly implement systematics and import them into the workspace
        self.implement_systematics(ws_factory)

        # work on remaining common variables and functions to be implemented
        self.import_expressions(ws_factory, items_lowpriority)
        
        if self.debug_mode:
            ws_factory.Print()
        
        # implement the scale factors on each sample
        self.implement_samples(ws_factory)
        
        if self.debug_mode:
            ws_factory.Print()
        
        # generate the combined signal + background pdf
        self.implement_sum_pdf(ws_factory)
        sum_pdf = ws_factory.pdf(SUM_PDF_NAME)

        self.implement_blind_range(ws_factory)
        
        self.check_nuisance_parameters(sum_pdf, core_argsets.nuis_set)
        
        # keep track of correlated variables
        correlated_terms = self.get_correlated_terms(ws_factory, core_argsets.nuis_set)
        correlated_str = ",".join(correlated_terms)
        
        self.stdout.debug(f"REGTEST: The following variables will not be renamed: {correlated_str}")
        # import pdf to new workspace, rename objects as appropriate
        self.import_object(ws, ws_factory.pdf(SUM_PDF_NAME),
                           ROOT.RooFit.RenameAllNodes(category),
                           ROOT.RooFit.RenameAllVariablesExcept(category, correlated_str))
        
        # import constraint terms
        self.attach_constraints(ws, core_argsets.constr_set)
        
        # define variable sets: observable, POIs, nuisance parameters, global observables
        poi_names = self.get_poi_names()
        core_argsets.poi_set = core_argsets.get_argset_by_names(ws, poi_names)
        core_argsets.define_argsets(ws)
        
        if self.debug_mode:
            ws.Print()
            
        obs = self.get_observable(ws)
        weight_var = ROOT.RooRealVar("wt", "wt", 1)
        dataset, datahist = self.generate_dataset(obs, weight_var)
    
        # prepare binned data
        dataset_binned_name = f"{OBS_DATASET_NAME}binned"
        dataset_binned = dataset.Clone(dataset_binned_name)
        
        # consider that blinded analysis the number of bins will reduce
        do_blind = self.core_config['do_blind']
        if do_blind:
            blind_range = self.cat_config[category]['dataset']['blind_range']
        else:
            blind_range = None
        blind_sf = 1
        if do_blind and blind_range:
            blind_sf = 1 - (blind_range[1] - blind_range[0]) / (obs.getMax() - obs.getMin())
        if (dataset.numEntries() > (obs.numBins() * blind_sf)):
            dataset_binned = ROOT.RooDataSet(dataset_binned_name, dataset_binned_name,
                                             ROOT.RooArgSet(obs, weight_var),
                                             ROOT.RooFit.WeightVar(weight_var))
            RooDataSet.fill_from_TH1(dataset_binned, datahist, blind_range=blind_range,
                                     min_bin_value=self.EPSILON / 1000.)
        num_events = pretty_value(dataset.sumEntries(), 5)
        self.stdout.info(f"REGTEST: Number of data events: {num_events}")
        self.import_object(ws, dataset)
        self.import_object(ws, dataset_binned)
        
        if self.core_config['generate_hist_data']:
            datahist_name = f"{OBS_DATASET_NAME}hist"
            rdatahist = ROOT.RooDataHist(datahist_name, datahist_name, ROOT.RooArgSet(obs), datahist)
            ws.Import(rdatahist)
        
        self.datahists[category]       = datahist
        self.datasets[category]        = dataset
        self.datasets_binned[category] = dataset_binned
        
        return ws
    
    def implement_sum_pdf(self, ws:ROOT.RooWorkspace):
        cat_config = self.get_active_category_config()
        samples = cat_config['samples']
        sum_pdf_components = [f"{sample.norm_name}*{sample.model_name}" for _, sample in samples.items()]
        sum_pdf_str = f"SUM::{SUM_PDF_NAME}({','.join(sum_pdf_components)})"
        self.stdout.debug(f"REGTEST: Sum pdf expr = {sum_pdf_str}")
        self.import_expression(ws, sum_pdf_str)
        
    def implement_samples(self, ws:ROOT.RooWorkspace):
        category = self.active_category
        cat_config = self.get_active_category_config()
        resp_func_map = cat_config['resp_func_map']
        samples = cat_config['samples']
        for _, sample in samples.items():
            self.stdout.debug(f'DEBUG: Implementing sample for the process "{sample.process}"')
            self.import_expressions(ws, sample.shape_factors)
            sample.norm_name = YIELD_PREFIX + sample.process
            norm_factor_components = []
            sample_norm_factors = self.import_expressions(ws, sample.norm_factors)
            norm_factor_components.append(sample_norm_factors)
            # common systematics: only add if exist
            for syst_domain in sample.syst_domains:
                if syst_domain in resp_func_map:
                    norm_factor_components.append(resp_func_map[syst_domain])
                # sample without systematics
                elif sample.process == syst_domain:
                    continue
                else:
                    raise RuntimeError(f"unknown systematic domain {syst_domain} in sample {sample.process}")
            norm_factor_components_str = ",".join(norm_factor_components)
            norm_factor_expr = f"prod::{sample.norm_name}({norm_factor_components_str})"
            self.import_expression(ws, norm_factor_expr)
            proc_yield = ws.function(sample.norm_name).getVal()
            self.stdout.info(f'INFO: Yield for category "{category}" '
                             f'process "{sample.process}": {proc_yield}')
            if self.debug_mode:
                ws.Print()
            self.implement_sample_model(ws, sample)      
    
    def implement_systematics(self, ws:ROOT.RooWorkspace):
        cat_config = self.get_active_category_config()
        systematics = cat_config['systematics']
        for domain, syst_list in systematics.items():
            syst_domain = SystematicsDomain(domain)
            for syst in syst_list:
                is_shape = syst.is_shape_syst()
                if syst_domain.is_shape is None:
                    syst_domain.is_shape = is_shape
                elif syst_domain.is_shape != is_shape:
                    raise RuntimeError(f'found mixture of shape and non-shape systematics in the domain "{domain}"')
                nuis_expr = f"{syst.name}[0, -5, 5]"
                nuis_name = self.import_expression(ws, nuis_expr)
                nuis = ws.var(nuis_name)
                uncert_hi_expr = syst.get_uncert_hi_expr()
                uncert_lo_expr = syst.get_uncert_lo_expr()
                uncert_hi_name = self.import_expression(ws, uncert_hi_expr)
                uncert_lo_name = self.import_expression(ws, uncert_lo_expr)
                uncert_hi = ws.var(uncert_hi_name)
                uncert_lo = ws.var(uncert_lo_name)
                syst_domain.add_item(nuis, syst.nominal, uncert_hi, uncert_lo,
                                     syst.get_interp_code(), syst.constr_term)
            self.build_response_function(ws, syst_domain)
        
    def generate_dataset(self, obs:ROOT.RooRealVar, weight_var:ROOT.RooRealVar):
        category_name = self.active_category
        cat_config = self.get_active_category_config()
        data_config = cat_config['dataset']
        obs_min = obs.getMin()
        obs_max = obs.getMax()
        n_bins = obs.getBins()
        do_blind = self.core_config['do_blind']
        if do_blind:
            blind_range = cat_config['dataset']['blind_range']
        else:
            blind_range = None
        data_scale = data_config['scale_data']
        data_name = OBS_DATASET_NAME
        filename = data_config['filename']
        filepath = self.path_manager.get_relpath(filename)
        filetype = data_config['filetype']
        obs_plus_weight = ROOT.RooArgSet(obs, weight_var)
        
        dataset = ROOT.RooDataSet(data_name, data_name, obs_plus_weight,
                                  ROOT.RooFit.WeightVar(weight_var))
        datahist = ROOT.TH1D(category_name, category_name, n_bins, obs_min, obs_max)
        datahist.Sumw2()
        
        is_counting = cat_config['category_type'] == DATA_SOURCE_COUNTING
        num_data = data_config['num_data']
        
        # counting experiment
        if (is_counting and num_data >= 0):
            bin_center = (obs_min + obs_max) / 2.
            if num_data == 0:
                weight = self.EPSILON / 1000.
            else:
                weight = num_data * data_scale
            obs.setVal(bin_center)
            weight_var.setVal(weight)
            dataset.add(ROOT.RooArgSet(obs, weight_var), weight)
            datahist.Fill(bin_center, weight)
        #  histogram
        elif (filetype == DATA_SOURCE_HISTOGRAM):
            histname = data_config['histname']
            h = self.get_histogram(filename, histname)
            neg_bin_indices, _ = TH1.remove_negative_bins(h)
            for i in neg_bin_indices:
                self.stdout.warning(f"REGTEST: Input data histogram bin {i} in "
                                    f"category {category_name} has negative weight. "
                                    f"Will force it to zero.", "red")
            if blind_range:
                TH1.apply_blind_range(h, blind_range)
            RooDataSet.fill_from_TH1(dataset, h,
                                     blind_range=blind_range,
                                     min_bin_value=self.EPSILON / 1000.,
                                     weight_scale=data_scale)
            rebin_ratio = TH1.rebin_to_variable(h, obs, tol=self.EPSILON)
            if rebin_ratio != 1:
                self.stdout.warning(f"REGTEST: Rebinning input data histogram "
                                    f"`{histname}` in category {category_name} by "
                                    f"{rebin_ratio} to match the binning of the observable")
            TH1.copy_histogram_in_effective_range(h, datahist, tol=self.EPSILON,
                                                  weight_scale=data_scale)
        else:
            if filetype == DATA_SOURCE_ASCII:
                filepath = self.path_manager.get_relpath(filename)
                dataset_tmp = self.get_dataset_from_ascii_file(filepath, obs, weight_var)
            else:
                cut      = data_config['cut']
                treename = data_config['treename']
                varname  = data_config['varname']
                chain, _ = self.get_chain_and_branch(filepath, treename, varname, cut)
                x_tree = ROOT.RooRealVar(varname, varname, obs.getMin(), obs.getMax())
                _dataset_name = f"{OBS_DATASET_NAME}_tmp"
                dataset_tmp = ROOT.RooDataSet(_dataset_name, _dataset_name,
                                              ROOT.RooArgSet(x_tree),
                                              ROOT.RooFit.Import(chain))
            xdata_tmp = dataset_tmp.get().first()          
            for i in range(dataset_tmp.numEntries()):
                dataset_tmp.get(i)
                x_val = xdata_tmp.getVal()
                obs.setVal(x_val)
                weight = dataset_tmp.weight() * data_scale
                weight_var.setVal(weight)
                # ignore data in the blind range
                if (blind_range) and (x_val > blind_range[0]) and (x_val < blind_range[1]):
                    continue
                dataset.add(ROOT.RooArgSet(obs, weight_var), weight)
                datahist.Fill(x_val, weight)
        n_event = dataset.sumEntries()
        n_event_binned = datahist.Integral()
        if abs(n_event - n_event_binned) > self.EPSILON:
            raise RuntimeError(f"Binned ({n_event_binned}) and unbinned ({n_event}) "
                               f"datasets have different number of entries in "
                               f"category {category_name}")
        if self.debug_mode:
            dataset.Print("v")
        # inject ghost
        if data_config['inject_ghost']:
            RooDataSet.add_ghost_weights(dataset, blind_range=blind_range, ghost_weight=self.EPSILON / 1000.)
            TH1.add_ghost_weights(datahist, blind_range=blind_range, ghost_weight=self.EPSILON / 1000.)
        return dataset, datahist
    
    def get_dataset_from_ascii_file(self, file:str, obs:ROOT.RooRealVar, weight_var:ROOT.RooRealVar):
        cat_config = self.get_active_category_config()
        weighted_data = cat_config['weighted_data']
        if weighted_data:
            dataset_name = f"{OBS_DATASET_NAME}_tmp"
            dataset = ROOT.RooDataSet(dataset_name, dataset_name,
                                      ROOT.RooArgSet(obs, weight_var),
                                      ROOT.RooFit.WeightVar(weight_var))
            try:
                data = np.loadtxt(file)
            except:
                raise RuntimeError('invalid ascii file format')
            assert (data.ndim == 2) and (data.shape[1] == 2)
            for data_i in data:
                obs.setVal(data_i[0])
                weight.setVal(data_i[1])
                dataset.add(ROOT.RooArgSet(obs, weight), weight)
        else:
            dataset = ROOT.RooDataSet.read(file, ROOT.RooArgList(obs))
        return dataset    
        
    def release_ghost(self, dataset:ROOT.RooDataSet, datahist:ROOT.RooDataHist,
                      obs:ROOT.RooRealVar, weight_var:ROOT.RooRealVar, ghost_weight:float):
        do_blind = self.core_config['do_blind']
        cat_config = self.get_active_category_config()
        blind_range = cat_config['dataset']['blind_range']
        n_bins = datahist.GetNbinsX()
        for i in range(1, n_bins + 1):
            if datahist.GetBinContent(i) == 0:
                bin_center = datahist.GetBinCenter(i)
                if (do_blind and blind_range) and \
                   (bin_center > blind_range[0]) and (bin_center < blind_range[1]):
                    continue
                obs.setVal(bin_center)
                weight_var.setVal(ghost_weight)
                dataset.add(ROOT.RooArgSet(obs, weight_var), ghost_weight)
                # also put the ghost weight in the histograms
                datahist.SetBinContent(i, ghost_weight)
                
    def attach_constraints(self, ws:ROOT.RooWorkspace, constraints:ROOT.RooArgSet):
        category_name = self.active_category
        self.stdout.debug(f"REGTEST: Attaching contraint pdfs to workspace for the category {category_name}")
        sum_pdf_name = f"{SUM_PDF_NAME}_{category_name}"
        final_pdf_name = f"{FINAL_PDF_NAME}_{category_name}"
        remove_set = ROOT.RooArgSet()
        pdf = ws.pdf(sum_pdf_name)
        if not pdf:
            raise RuntimeError(f"sum pdf `{sum_pdf_name}` not defined")
        for constraint in constraints:
            constr_name = constraint.GetName()
            np_name = constr_name.replace(CONSTRTERM_PREFIX, "")
            if not pdf.getVariables().find(np_name):
                self.stdout.warning(f"WARNING: constraint term {constr_name} with NP "
                                    f"{np_name} is redundant in the category {category_name}. "
                                    f"It will be removed...", "red")
                remove_set.add(constraint)
            else:
                self.import_object(ws, constraint)
        for constraint in remove_set:
            constraints.remove(constraint)
        pdf_components = [sum_pdf_name] + [constraint.GetName() for constraint in constraints]
        pdf_str = ",".join(pdf_components)
        model_expr = f"PROD::{final_pdf_name}({pdf_str})"
        self.import_expression(ws, model_expr)
        
    def get_correlated_terms(self, ws:ROOT.RooWorkspace, nuis:ROOT.RooArgSet) -> List[str]:
        correlated_terms = []
        cat_config = self.get_active_category_config()
        # correlate custom variables
        correlated_terms.extend(cat_config['correlate_items'])
        # correlate POIs
        poi_names = self.get_poi_names()
        correlated_terms.extend(poi_names)
        # correlate nuisance parameters
        nuis_names = [np.GetName() for np in nuis]
        correlated_terms.extend(nuis_names)
        # correlate observable
        observable_name = cat_config['observable']['name']
        correlated_terms.append(observable_name)
        # remove duplicate terms
        correlated_terms = remove_list_duplicates(correlated_terms)
        # check validity of variables
        valid_correlated_terms = []
        for item in correlated_terms:
            # does not exist
            if (not ws.obj(item)):
                continue
            if (not ws.var(item)):
                raise RuntimeError(f"correlated variable `{item}` is not properly implemented"
                                   f" as RooRealVar in the workspace")
            valid_correlated_terms.append(item)
        return valid_correlated_terms
    
    def get_poi_names(self):
        return self.core_config['poi_names']
    
    def get_observable(self, ws:ROOT.RooWorkspace):
        cat_config = self.get_active_category_config()
        observable_name = cat_config['observable']['name']
        obs = ws.var(observable_name)
        if not obs:
            category = self.active_category
            raise RuntimeError(f"observable for the category `{category}` not defined in the workspace")
        return obs

    def implement_blind_range(self, ws:ROOT.RooWorkspace):
        # range_name = "" : fully blinded
        # range_name = None: no blinding
        # otherwise partially blinded
        cat_config = self.get_active_category_config()
        category = self.active_category
        obs = self.get_observable(ws)
        obs_range = [obs.getMin(), obs.getMax()]
        # now handle blinding if needed
        do_blind = self.core_config["do_blind"]
        if do_blind:
            blind_range = cat_config['dataset']['blind_range']
            if blind_range:
                self.stdout.info(f"REGTEST: Implement blind range [{blind_range[0]}, {blind_range[1]}]")
                sideband_lo_range_name = f"{RANGE_NAME_SB_LO}_{category}"
                sideband_hi_range_name = f"{RANGE_NAME_SB_HI}_{category}"
                blind_range_name = f"{RANGE_NAME_BLIND}_{category}"
                obs.setRange(sideband_lo_range_name, obs_range[0], blind_range[0])
                obs.setRange(blind_range_name, blind_range[0], blind_range[1])
                obs.setRange(sideband_hi_range_name, blind_range[1], obs_range[1])
                if (blind_range[0] == obs_range[0]) and \
                   (blind_range[1] == obs_range[1]):
                    self.stdout.info(f"REGTEST: Category `{category}` is fully blinded. "
                                     f"No side-band exists.")
                    cat_config['range_name'] = ""
                elif (blind_range[1] == obs_range[1]):
                    cat_config['range_name'] = sideband_lo_range_name
                elif (blind_range[0] == obs_range[0]):
                    cat_config['range_name'] = sideband_hi_range_name
                else:
                    cat_config['range_name'] = f"{sideband_lo_range_name},{sideband_hi_range_name}"
            # do blind but no range specified : fully blinded
            else:
                cat_config['range_name'] = ""
        else:
            cat_config['range_name'] = None
        
    @staticmethod
    def _generate_expr(head:str, argset:ROOT.RooArgSet, close_expr:bool=True):
        tokens = [i.GetName() for i in argset]
        expr_str = head + ",".join(tokens)
        if close_expr:
            expr_str += ")"
            expr_str = expr_str.strip().replace(",)", ")")
        return expr_str
    
    def _fix_item_priority(self):
        cat_config = self.get_active_category_config()
        items_highpriority  = cat_config['items_highpriority']
        items_lowpriority   = cat_config['items_lowpriority']
        _items_highpriority = []
        for item in items_highpriority:
            if self._item_has_dependent(item, items_lowpriority):
                items_lowpriority.append(item)
            else:
                _items_highpriority.append(item)
        cat_config['items_highpriority'] = _items_highpriority
    
    def _item_has_dependent(self, expr:str, reference_list:List[str]):
        _, item_type = self._get_object_name_and_type_from_expr(expr)
        self.stdout.debug(f"REGTEST: item `{expr}` has type `{item_type}`")
        if item_type in ['variable', 'defined']:
            return False
        item_list = self._decompose_function_expr(expr)
        self.stdout.debug(f"REGTEST: Composition of high priority item {expr}: "
                          f"{', '.join(item_list)}")
        for this_name in item_list:
            for reference in reference_list:
                that_name, _ = self._get_object_name_and_type_from_expr(reference)
                if this_name == that_name:
                    return True
        return False

    @staticmethod
    def _decompose_function_expr(expr:str):
        start = expr.find("(")
        end = expr.rfind(")")
        if (start < 0) or (end < 0):
            raise RuntimeError(f"not a valid function expression `{expr}`")
        tokens = expr[start + 1 : end].split(",")
        if ("expr::" in expr) or ("EXPR::" in expr):
            tokens = tokens[1:]
        return tokens
        
    @staticmethod
    def _get_object_name_and_type_from_expr(expr:str):
        if ("::" in expr) and ("(" in expr):
            object_type = "function"
            object_name = expr.split("::")[1].split("(")[0]
        elif ("[" in expr):
            object_type = "variable"
            object_name = expr.split("[")[0]
        elif (":" in expr) and not ("::" in expr):
            raise RuntimeError(f"syntax error for the expression `{expr}`: missing colon pair")
        else:
            object_type = "defined"
            object_name = expr
        return object_name, object_type
            
    def import_uncert_expr(self, ws:ROOT.RooWorkspace, expr:str, var_name:str, prefix:str):
        if is_float(expr):
            new_expr = f"{prefix}{var_name}[{expr}]"
            return self.import_expression(ws, new_expr)
        else:
            return self.import_expression(ws, expr)

    def _translate_keyword(self, expr:str):
        category_name = self.active_category
        observable_name = self.cat_config[category_name]['observable']['name']
        expr = expr.replace(RESPONSE_KEYWORD, 
                            RESPONSE_PREFIX + \
                            SHAPE_SYST_KEYWORD + "_")
        expr = expr.replace(OBSERVABLE_KEYWORD,
                            observable_name)
        expr = expr.replace(CATEGORY_KEYWORD,
                            category_name)
        expr = expr.replace(COMMON_SYST_DOMAIN_KEYWORD,
                            Systematic.common_domain())
        expr = expr.replace(LT_OP, "<")
        expr = expr.replace(LE_OP, "<=")
        expr = expr.replace(GT_OP, ">")
        expr = expr.replace(GE_OP, ">=")
        expr = expr.replace(AND_OP, "%%")
        expr = expr.replace(OR_OP, "||")
        return expr
       
    def _translate_node_attrib(self, node:Dict, attrib_name:str, process:str="",
                               required:bool=True, default:str=""):
        expr = self._get_node_attrib(node, attrib_name, required, default)
        expr = self._translate_keyword(expr)
        if (PROCESS_KEYWORD in expr):
            if process == "":
                raise RuntimeError(f"process name not provided for the expression `{expr}`")
            expr = expr.replace(PROCESS_KEYWORD, f"_{process}")
        return expr
    
    def get_chain_and_branch(self, filename:str, treename:str, varname:str,
                             cut:str="", check_only:bool=False):
        chain = ROOT.TChain(treename)
        filenames = filename.split(",")
        filepaths = [self.path_manager.get_relpath(filename) for filename in filenames]
        for filepath in filepaths:
            if not os.path.exists(filepath):
                raise FileNotFoundError(f"input data file `{filepath}` not found")
            status = chain.AddFile(filepath, -1)
            if not status:
                raise RuntimeErrro(f"cannot find TTree `{treename}` in data file `{filepath}`")
        if cut != "":
            chain = chain.CopyTree(cut)
        branch = chain.FindBranch(varname)
        if not branch:
            raise RuntimeError(f"cannot find TBranch `{varname}` in TTree `{treename}` "
                               f"in data file `{filename}`")
        if check_only:
            return Nones
        return chain, branch
    
    def get_histogram(self, filename:str, histname:str, check_only:bool=False):
        filepath = self.path_manager.get_relpath(filename)
        if not os.path.exists(filepath):
            raise FileNotFoundError(f"input data file `{filepath}` not found")
        f = ROOT.TFile(filepath)
        category_name = self.active_category
        data_config = self.cat_config[category_name]['dataset']
        histname = data_config['histname']
        histogram = f.Get(histname)
        if not isinstance(histogram, ROOT.TH1):
            raise RuntimeError(f"no TH1 named `{histanme}` found in data file `{filepath}`")
        if check_only:
            return None
        histogram.SetDirectory(0)        
        return histogram
    
    def data_file_sanity_check(self):
        category_name = self.active_category
        category_type = self.cat_config[category_name]['category_type']
        data_config = self.cat_config[category_name]['dataset']
        filename = data_config['filename']    
        if category_type == DATA_SOURCE_COUNTING:
            if (filename != "") and (self.check_file_exist(filename)):
                raise FileNotFoundError(f"input data file `{filepath}` not found")
            if (filename == "") and (data_config.get("num_data", -1) < 0):
                raise RuntimeError("pleasee provide either a valid input data filename, or "
                                   "a valid number of data events for a counting experiment")
        else:
            if data_config['filetype'] != DATA_SOURCE_ASCII:
                if data_config['filetype'] == DATA_SOURCE_HISTOGRAM:
                    histname = data_config['histname']
                    self.get_histogram(filename, histname, check_only=True)
                else:
                    treename = data_config['treename']
                    varname = data_config['varname']
                    self.get_chain_and_branch(filename, treename, varname, check_only=True)
    
    def fetch_node(self, nodes:List, tag:str, xml_filename:str, xml_type:str="channel", 
                   allow_missing:bool=False, allow_multiple:bool=False):
        target_node = [node for node in nodes if node['tag'] == tag]
        if (len(target_node) > 1) and (not allow_multiple):
            raise RuntimError(f"Multiple `{tag}` node found in {xml_type} XML file `{xml_filename}`")
        elif (len(target_node) == 0) and (not allow_missing):
            raise RuntimError(f"No `{tag}` node found in {xml_type} XML file `{xml_filename}`")
        elif (len(target_node) == 1) and (not allow_multiple):
            target_node = target_node[0]
        return target_node
    
    def get_active_category_config(self):
        category_name = self.active_category
        if category_name not in self.cat_config:
            raise RuntimeError(f"no config info found for the category `{category_name}`")
        cat_config = self.cat_config[category_name]
        return cat_config
        
    def read_systematics_node(self, node:Dict, domain:str):
        # Read XML config file. Domain name is also passed in some cases, and will replace the :process: keyword in case it is there
        cat_config = self.get_active_category_config()
        syst_expr = self._get_node_attrib(node, "Name")
        self.stdout.debug(f"REGTEST: Reading systematic `{syst_expr}`")
        syst = Systematic()
        syst.name = self._translate_node_attrib(node, "Name", domain)
        # if the process of the systematic is specified, use the 
        # specified process. Otherwise it is set to empty
        syst.process = self._translate_node_attrib(node, "Process", domain, False, "")
        syst.whereto = self._get_node_attrib(node, "WhereTo")
        syst.nominal = self._get_node_attrib(node, "CentralValue", dtype="float")
        syst.constr_term = self._get_node_attrib(node, "Constr")
        syst.set_domain(domain)
        syst.validate()
        magitude_expr = self._get_node_attrib(node, "Mag", dtype="str_list")
        syst.set_magnitudes(magitude_expr)
        self.add_syst_object(syst)

    def add_syst_object(self, syst:Systematic):
        cat_config = self.get_active_category_config()
        systematics = cat_config['systematics']
        if syst.domain not in systematics:
            systematics[syst.domain] = []
        syst_list = systematics[syst.domain]
        if any(syst_i == syst for syst_i in syst_list):
            raise RuntimeError(f"systematic `{syst.name}` applied on `{syst.whereto}` "
                               f"is duplicated for the process `{syst.process}`")
        syst_list.append(syst)
        
    def add_sample_object(self, sample:Sample):
        cat_config = self.get_active_category_config()
        sample_map = cat_config['samples']
        process = sample.process
        if process in sample_map:
            category = self.active_category
            raise RuntimeError(f"duplicated sample `{sample.process}` in the category `{category}`")
        sample_map[process] = sample
        
    def read_item_node(self, node:Dict):
        cat_config = self.get_active_category_config()
        item = self._translate_node_attrib(node, "Name")
        if (RESPONSE_KEYWORD in item) or \
           (RESPONSE_PREFIX in item):
            cat_config['items_lowpriority'].append(item)
        else:
            cat_config['items_highpriority'].append(item)
        
    def read_channel_import_node(self, node:Dict, ws:ROOT.RooWorkspace):
        filename = self._translate_node_attrib(node, "FileName")
        filepath = self.path_manager.get_relpath(filename)
        if not os.path.exists(filepath):
            raise FileNotFoundError(f"import XML file `{filepath}` not found")
        self.stdout.debug(f"REGTEST: Reading import XML file `{filepath}`")
        xml_data = TXMLTree.load_as_dict(filepath)
        nodes = xml_data['children']
        self.read_channel_xml_nodes(nodes, ws)
        
    def read_sample_import_node(self, node:Dict, sample:Sample):
        filename = self._translate_node_attrib(node, "FileName")
        filepath = self.path_manager.get_relpath(filename)
        if not os.path.exists(filepath):
            raise FileNotFoundError(f"import XML file `{filepath}` not found")
        self.stdout.info(f"REGTEST: Importing items for `{sample.process}` from {filepath}")
        xml_data = TXMLTree.load_as_dict(filepath)
        nodes = xml_data['children']
        self.read_sample_subnode(nodes, sample)
    
    def read_sample_factor_node(self, node:Dict, sample:Sample):
        """Parse Norm/Shape factors from a Sample Node
        
        Example implementation:
        <Sample Name="ggF" InputFile="models/ggF.xml" ImportSyst=":common:">
          <NormFactor Name="yield_ggF[0.0078]" Correlate=0 />
        </Sample>
        """
        factor_type = SampleFactorType.parse(node["tag"])
        factor = self._translate_node_attrib(node, "Name", sample.process)
        if factor_type == SampleFactorType.NormFactor:
            sample.norm_factors.append(factor)
        elif factor_type == SampleFactorType.ShapeFactor:
            sample.shape_factors.append(factor)
        is_correlated = self._get_node_attrib(node, "Correlate", required=False, default=0, dtype="int")
        if is_correlated:
            cat_config = self.get_active_category_config()
            item_name, item_type = self._get_object_name_and_type_from_expr(factor)
            cat_config['correlate_items'].append(item_name)
    
    def read_data_node(self, node:Dict, ws:ROOT.RooWorkspace):
        cat_config = self.get_active_category_config()
        # extract observable information
        observable_expr = self._translate_node_attrib(node, "Observable")
        observable_name = self.import_expression(ws, observable_expr)
        observable = ws.var(observable_name)
        observable_min = observable.getMin()
        observable_max = observable.getMax()
        if observable_min >= observable_max:
            raise RuntimeError(f"invalid range for observable `{observable_name}`: "
                               f"min={observable_min}, max={observable_max}")
        is_counting = cat_config['category_type'] == DATA_SOURCE_COUNTING
        if is_counting:
            observable_nbins = 1
        else:
            observable_nbins = int(self._get_node_attrib(node, "Binning"))
        ws.var(observable_name).setBins(observable_nbins)
        
        cat_config['observable'] = {
            'name': observable_name,
            'min': observable_min,
            'max': observable_max,
            'nbins': observable_nbins
        }
        
        core_argsets = cat_config['core_argsets']
        core_argsets.obs_set = ROOT.RooArgSet(observable)
        
        data_filename = self._translate_node_attrib(node, "InputFile", "",
                                                    required=not is_counting,
                                                    default="")
        data_filetype = self._get_node_attrib(node, "FileType", required=False,
                                              default=DATA_SOURCE_ASCII).lower()
        data_config = {}
        data_config['filename'] = data_filename
        data_config['filetype'] = data_filetype
        # reading from text file (event-by-event)
        if data_filetype == DATA_SOURCE_ASCII:
            cat_config['weighted_data'] = self._get_node_attrib(node, "HasWeight", required=False,
                                                                default=False, dtype="bool")
        # reading from root histogram
        elif data_filetype == DATA_SOURCE_HISTOGRAM:
            data_config['histname'] = self._translate_node_attrib(node, ["HistName", "HistoName"])
        # reading from root ntuples
        else:
            data_config['treename'] = self._translate_node_attrib(node, "TreeName")
            data_config['varname'] = self._translate_node_attrib(node, "VarName")
            data_config['cut'] = self._translate_node_attrib(node, "Cut", "",
                                                             required=False, default="")

        inject_ghost = self._get_node_attrib(node, "InjectGhost", required=False, default=0, dtype="int")
        data_config['inject_ghost'] = inject_ghost
        if (is_counting and data_filename == ""):
            data_config['num_data'] = self._get_node_attrib(node, "NumData", dtype="int")
        else:
            data_config['num_data'] = -1
        data_config['scale_data'] = self._get_node_attrib(node, "ScaleData",
                                                          required=False, default=1, dtype="float")
        if self.core_config['scale_lumi'] > 0:
            data_config['scale_data'] *= self.core_config['scale_lumi']
        blind_range = self._get_node_attrib(node, "BlindRange", required=False, default=None)
        if blind_range:
            blind_range = blind_range.split(",")
            if (len(blind_range) != 2) or (not is_float(blind_range[0])) or (not is_float(blind_range[1])):
                raise RuntimeError(f"invalid blind range format: {blind_range}")
            blind_range = [float(r) for r in blind_range]
            if (blind_range[1] <= blind_range[0]) or (blind_range[0] < observable_min) or (blind_range[1] > observable_max):
                raise RuntimeError(f"invalid blind range given: min={blind_range[0]}, max={blind_range[1]}")
        data_config['blind_range'] = blind_range
        cat_config['dataset'] = data_config
        
        self.data_file_sanity_check()
        
    def read_correlate_node(self, node:Optional[Dict]=None):
        cat_config = self.get_active_category_config()
        if node:
            item_str = node['text']
            if not item_str:
                filename = cat_config['input_file']
                raise RuntimeError(f"no items defined for correlation in channel XML file `{filename}`")
            correlate_items = item_str.split(",")
        else:
            correlate_items = []
        cat_config['correlate_items'] = correlate_items
        
    def read_sample_node(self, node:Dict):
        cat_config = self.get_active_category_config()
        is_counting = cat_config['category_type'] == DATA_SOURCE_COUNTING
        sample = Sample()
        sample.process = self._translate_node_attrib(node, "Name")
        sample.input_file = self._translate_node_attrib(node, "InputFile", sample.process,
                                                        required=not is_counting, default="")
        syst_domains = self._translate_node_attrib(node, "ImportSyst", sample.process,
                                                   required=False, default=SELF_SYST_DOMAIN_KEYWORD)
        syst_domains = re.sub('\s+', '', syst_domains).split(",")
        sample.set_syst_domains(syst_domains)
        
        # assemble the yield central value
        
        if cat_config['luminosity'] > 0:
            multiply_lumi = self._get_node_attrib(node, "MultiplyLumi",
                                                  required=False, default=1, dtype="int")
            if multiply_lumi:
                sample.norm_factors.append(LUMI_NAME)

        norm_factor_names = {
            'norm': ('Norm', NORM_PREFIX),
            'xs': ('XSection', XS_PREFIX),
            'br': ('BR', BR_PREFIX),
            'efficiency': ('SelectionEff', EFFICIENCY_PREFIX),
            'acceptance': ('Acceptance', ACCEPTANCE_PREFIX),
            'correction': ('Correction', CORRECTION_PREFIX)
        }
        
        # sample level normalization factors
        for key, (name, prefix) in norm_factor_names.items():
            norm_value = self._translate_node_attrib(node, name, sample.process,
                                                     required=False, default="")
            if norm_value != "":
                sample.norm_factors.append(f"{prefix}_{sample.process}[{norm_value}]")
                
        sample.share_pdf_group = self._translate_node_attrib(node, "SharePdf", sample.process,
                                                             required=False, default="")
        subnodes = node['children']
        self.stdout.debug(f"REGTEST: Reading sample `{sample.process}`")
        self.read_sample_subnode(subnodes, sample)    
        self.add_sample_object(sample)
        
    def read_sample_subnode(self, nodes:List, sample:Sample):
        for node in nodes:
            name = node['tag']
            if name == "Systematic":
                self.read_systematics_node(node, sample.process)
            elif SampleFactorType.has_member(name):
                self.read_sample_factor_node(node, sample)
            elif name == "ImportItems":
                self.read_sample_import_node(node, sample)
            else:
                raise RuntimeError(f"unsupported node `{name}`")        
    
    def read_channel_xml_nodes(self, nodes:List, ws:ROOT.RooWorkspace):
        for node in nodes:
            name = node['tag']
            if name == "Item":
                self.read_item_node(node)
            elif name == "Systematic":
                self.read_systematics_node(node, domain=Systematic.default_domain())
            elif name == "Sample":
                self.read_sample_node(node)
            elif name == "ImportItems":
                self.read_channel_import_node(node, ws)
            else:
                raise RuntimeError(f"unsupported node `{name}`")
                
    def build_response_function(self, ws:ROOT.RooWorkspace, syst_domain:SystematicsDomain):
        cat_config = self.get_active_category_config()
        core_argsets = cat_config['core_argsets']
        resp_func_map = cat_config['resp_func_map']
        resp_func = syst_domain.get_response_function()
        ws.Import(resp_func, ROOT.RooFit.RecycleConflictNodes())
        if (not syst_domain.is_shape):
            resp_func_map[syst_domain.domain] = syst_domain.get_response_name()
        for i in range(syst_domain.nuis_list.size()):
            np_name, glob_name, constr_name = syst_domain.get_np_glob_constr_names(i)
            if syst_domain.constr_term_list[i] == CONSTR_DFD:
                self.stdout.debug(f"REGTEST: Creating DFD constraint term for NP {np_name}")
                constr_expr = (f"EXPR::{constr_name}('1/((1+exp(@2*(@0-@3-@1)))*(1+exp(-1*@2*(@0-@3+@1))))', "
                               f"{np_name}, DFD_e[1], DFD_w[500], {glob_name}[0,-5,5])")
                self.import_expression(ws, constr_expr, True)
            else:
                self.stdout.debug(f"REGTEST: Creating RooGaussian constraint term for NP {np_name}")
                constr_expr = f"RooGaussian::{constr_name}({np_name},{glob_name}[0,-5,5],1)"
            self.import_expression(ws, constr_expr, True)
            core_argsets.nuis_set.add(ws.var(np_name), True)
            core_argsets.glob_set.add(ws.var(glob_name), True)
            core_argsets.constr_set.add(ws.pdf(constr_name), True)
            self.stdout.debug(f"REGTEST: Finished implementing the systematic {np_name}")
            
    def implement_sample_model(self, ws:ROOT.RooWorkspace, sample:Sample):
        self.stdout.debug("REGTEST: Getting model")
        sample.update_model_name()
        if ws.pdf(sample.model_name):
            if is_shared_pdf:
                # use shared pdf
                self.stdout.debug(f"REGTEST: PDF {sample.model_name} has been created in the workpace.")
                return None
            else:
                raise RuntimeError(f"PDF {sample.model_name} already exists but the user asks to create it again")
        self.stdout.debug(sample.model_name)
        cat_config = self.get_active_category_config()
        is_counting = cat_config['category_type'] == DATA_SOURCE_COUNTING
        if is_counting:
            # in counting experiment we only need a uniform pdf
            obs_name = cat_config['observable']['name']
            self.import_expression(ws, f"RooUniform::{sample.model_name}({obs_name})")
        else:
            filename = sample.input_file            
            self.read_model_xml(filename, ws, sample)
            
    def read_model_xml(self, filename:str, ws:ROOT.RooWorkspace, sample:Sample):
        cat_config = self.get_active_category_config()
        filepath = self.path_manager.get_relpath(filename)
        if not os.path.exists(filepath):
            raise FileNotFoundError(f'model xml file "{filepath}" does not exist')
        root_node = TXMLTree.load_as_dict(filepath)
        
        model_type = self._get_node_attrib(root_node, "Type").lower()
        if model_type == USERDEF_MODEL:
            self.stdout.debug(f'REGTEST: Creating user-defined pdf from "{filepath}"')
            self.build_userdef_model(root_node, ws, sample)
        elif model_type == EXTERNAL_MODEL:
            self.build_external_model(root_node, ws, sample)
        elif model_type == HISTOGRAM_MODEL:
            self.build_histogram_model(root_node, ws, sample)
        else:
            raise RuntimeError(f"unknown model type: {model_type} . Available choice: "
                               f" {USERDEF_MODEL}, {EXTERNAL_MODEL}, "
                               f" {HISTOGRAM_MODEL}")
    
    def build_userdef_model(self, root_node:Dict, ws:ROOT.RooWorkspace, sample:Sample):
        cat_config = self.get_active_category_config()
        tag_name = sample.get_tag_name()
        obs_name = cat_config['observable']['name']
        cache_binning = self._get_node_attrib(root_node, "CacheBinning", False, "-1", dtype="int")
        obs = ws.var(obs_name)
        if cache_binning > 0:
            if not obs:
                raise RuntimeError("observable not initialized")
            # for the accuracy of Fourier transformation
            obs.setBins(cache_binning, "cache")
        is_success = False
        nodes = root_node['children']
        for node in nodes:
            name = node['tag']
            self.stdout.debug(f"REGTEST: Reading node `{name}`")
            if name == "Item":
                expr = self._translate_node_attrib(node, "Name", tag_name)
                self.import_expression(ws, expr)
            elif name == "ModelItem":
                factory_str = self._translate_node_attrib(node, "Name", tag_name)
                str_start = factory_str.find("::")+2
                str_end   = factory_str.find("(")
                pdf_name = factory_str[str_start:str_end]
                factory_str = factory_str[:str_start] + sample.model_name + factory_str[str_end:]
                self.import_expression(ws, factory_str)
                is_success = True
                break
            else:
                raise RuntimeError(f'unknown node "{name}"')
        if not is_success:
            raise RuntimeError('missing node "ModelItem" from model xml')
    
    def build_external_model(self, root_node:Dict, ws:ROOT.RooWorkspace, sample:Sample,
                             nuis:ROOT.RooArgSet, constraints:ROOT.RooArgSet, globobs:ROOT.RooArgSet):
        cat_config = self.get_active_category_config()
        _observable_name = cat_config['observable']['name']
        filename = self._translate_node_attrib(root_node, "Input")
        filepath = self.path_manager.get_relpath(filename)
        if not os.path.exists(filepath):
            raise FileNotFoundError(f'input file "{filepath}" not found')
        ws_name = self._translate_node_attrib(root_node, "WSName")
        model_name = self._translate_node_attrib(root_node, "ModelName")
        observable_name = self._translate_node_attrib(root_node, "ObservableName")
        tag_name = sample.get_tag_name()
        nodes = root_node['children']
        np_list, glob_list, constr_list = [], [], []
        self.stdout.debug(f"REGTEST: Use existing PDF `{model_name}` from the workspace `{filename}`")
        
        f_ext = ROOT.TFile(filename)
        ws_ext = f_ext.Get(ws_name)
        pdf_ext = ws_ext.function(model_name)
        
        name_maps = {}
        do_not_touch = [observable_name]
        
        for node in nodes:
            name = node['tag']
            if name == "Item":
                expr = self._translate_node_attrib(node, "Name", tag_name)
                self.import_object(ws, expr)
            elif name == "Fix":
                var_name = self._translate_node_attrib(node, "Name")
                var = ws_ext.var(var_name)
                if not var:
                    raise RuntimeError(f"no variable named `{var_name}` found im the workspace `{filepath}`")
                var.setConstant(1)
                var_val = self._get_node_attrib(node, "Value", False, "")
                if var_val != "":
                    var.setVal(float(var))
                self.stdout.debug(f"REGTEST: Fixed variable `{var_name}` at `{var.getVal()}")
            elif name == "Rename":
                # rename the object names in the input workspace
                old_name = self._translate_node_attrib(node, "OldName", tag_name)
                new_name = self._translate_node_attrib(node, "NewName", tag_name)
                do_not_touch.append(old_name)
                name_maps[old_name] = new_name
            elif name == "ExtSyst":
                # adding external systematics
                np_name = self._translate_node_attrib(node, "NPName")
                glob_name = self._translate_node_attrib(node, "GOName", "", False, "")
                constr_name = self._translate_node_attrib(node, "ConstrName", "",
                                                          required=glob_name!="", default="")
                np_list.append(np_name)
                glob_list.append(glob_name)
                constr_list.append(constr_name)
                do_not_touch.append(np_name)
            else:
                raise RuntimeError(f"unknown node `{name}`")
        # remove duplicated variables
        do_not_touch = ",".join(remove_list_duplicates(do_not_touch))
    
        # rename everything in the model by adding a tag to it
        ws_temp = ROOT.RooWorkspace("ws_temp")
        self.import_object(ws, pdf_ext, ROOT.RooFit.RenameAllNodes(tag_name),
                           ROOT.RooFit.RenameAllVariablesExcept(tag_name, do_not_touch))
        #ws_temp.importClassCode()
        if self.debug_mode:
            ws_temp.Print()
            
        pdf = ws_temp.function(f"{model_name}_{tag_name}")
        obs_ext = ws_temp.var(observable_name)
        if not obs_ext:
            raise RuntimeError(f"observable `{observable_name}` not found in the workspace `{filepath}`")
        if not isinstance(pdf, ROOT.RooAbsPdf):
            self.stdout.warning(f"WARNING: The object {model_name} is not a p.d.f as seen from RoOFit")
            self.stdout.warning("REGTEST: Constructing RooRealSumPdf for the pdf from histFactory")
            dummy_pdf = ROOT.RooUniform(f"{sample.model_name}_dummpy_pdf", "for RooRealSumPdf construction",
                                        obs_ext)
            frac = ROOT.RooRealVar(f"{sample.model_name}_real_pdf_frac", "for RooRealSumPdf construction", 1)
            pdf = ROOT.RooRealSumPdf(sample.model_name, "from histFactory", ROOT.RooArgList(pdf, dummy_pdf),
                                     ROOT.RooArgList(frac))
        else:
            name_maps[pdf.GetName()] = sample.model_name
        
        # rename observable (this step should never happen for a HistFactory workspace)
        if observable_name != _observable_name:
            name_maps[observable_name] = _observable_name
        self.stdout.info("REGTEST: The following variables will be renamed:")
        old_str = ",".join(name_maps.keys())
        new_str = ",".join(name_maps.values())
        self.stdout.info(f"REGTEST: Old: {old_str}")
        self.stdout.info(f"REGTEST: New: {new_str}")
        
        self.import_object(ws, pdf, ROOT.RooFit.RenameVariable(old_str, new_str),
                           ROOT.RooFit.RecycleConflictNodes())
        # import constraint terms
        for np_name, glob_name, constr_name in zip(np_list, glob_list, constr_list):
            new_np_name = np_name
            # check wheterh NP has been renamed. If yes, update the NP name
            for old_name, new_name in name_maps.items():
                if np_name == old_name:
                    new_np_name == new_name
                    break
            if glob_name != "":
                if (not ws_ext.var(np_name)) or (not ws_ext.var(glob_name)) or (not ws_ext.pdf(constr_name)):
                    raise RuntimeError(f"constraint pdf {constr_name} with NP = {np_name} and "
                                       f"GlobObs = {glob_name} does not exist")
                new_constr_name = CONSTRTERM_PREFIX + new_np_name
                new_glob_name = GLOBALOBS_PREFIX + new_np_name
                constr_pdf = ws_ext.pdf(constr_name)
                old_str = ",".join([constr_name, glob_name, np_name])
                new_str = ",".join([new_constr_name, new_glob_name, new_np_name])
                self.import_object(ws, constr_pdf, 
                                   ROOT.RooFit.RenameVariable(old_str, new_str))
                self.stdout.debug(f"REGTEST: Import systematics {new_np_name} "
                                  f"with constraint term {new_constr_name}")
                constraints.add(ws.pdf(new_constr_name), True)
                globobs.add(ws.var(new_glob_name), True)
            ws.var(new_np_name).setConstant(False)
            nuis.add(ws.var(new_np_name), True)
    
    def build_histogram_model(self, root_node:Dict, ws:ROOT.RooWorkspace, sample:Sample):
        cat_config = self.get_active_category_config()
        obs_name = cat_config['observable']['name']
        filename = self._translate_node_attrib(root_node, "Input")
        filepath = self.path_manager.get_relpath(filename)
        histname = self._translate_node_attrib(root_node, "ModelName")
        rebin = self._get_node_attrib(root_node, "Rebin", False, "-1", dtype="int")
        self.stdout.debug(f"REGTEST: Creating histogram pdf from `{filepath}`")
        h = self.get_histogram(filename, histname)
        if rebin > 0:
            h.Rebin(rebin)
        obs = ws.var(obs_name)
        if not obs:
            raise RuntimeError("observable not initialized")
        h_data = ROOT.RooDataHist("hdata", "hdata", obs, h)
        h_pdf = ROOT.RooHistPdf(sample.model_name, sample.model_name, obs, h_data)
        self.import_object(ws, h_pdf)
    
    def check_nuisance_parameters(self, model:ROOT.RooAbsPdf, nuis:ROOT.RooArgSet):
        cat_config = self.get_active_category_config()
        obs_name = cat_config['observable']['name']        
        var_set = model.getVariables()
        float_set = ROOT.RooArgSet()
        for var in var_set:
            if (not var.isConstant()):
                if var.getMin() == var.getMax():
                    var_name = var.GetName()
                    self.stdout.info(f"REGTEST: Fixing {var_name} to constant as it has same upper and loer boundary")
                    var.setConstant(True)
                else:
                    float_set.add(var)
        poi_obs_names = self.core_config['poi_names'] + [obs_name]
        # reove POI and observable from float set
        for name in poi_obs_names:
            float_var = float_set.find(name)
            if float_var:
                float_set.remove(float_var)
        num_np = len(nuis)
        num_poi = len(self.core_config['poi_names'])
        for var in float_set:
            if not nuis.find(var):
                nuis.add(var)
                var_name = var.GetName()
                self.stdout.info(f"REGTEST: Adding {var_name} to the nuisance parameter set")

        num_float = len(float_set)
        if num_float < num_np:
            self.stdout.warning(f"WARNING: There supposed to be {num_poi + num_np} free parameters, "
                                f"but only seen {num_float} in the category {self.active_category}. "
                                f"This is in principle not an issue, but please make sure you "
                                f"understand what you are doing")
            if self.debug_mode:
                self.stdout.debug("REGTEST: All free parameters: ")
                float_set.Print()
                self.stdout.debug("REGTEST: All nuisance parameters: ")
                nuis.Print()
        else:
            self.stdout.info("REGTEST: Number of nuisance parameters looks good!")
  
    def get_pdf_name(self, category:str):
        return f"{FINAL_PDF_NAME}_{category}"
    
    def implement_model_config_and_sets(self, ws:"ROOT.RooWorkspace", mc:"ROOT.RooStats.ModelConfig",
                                        obs:ROOT.RooArgSet, pois:ROOT.RooArgSet,
                                        nuis:ROOT.RooArgSet, globs:ROOT.RooArgSet):
        ws.defineSet("nuisanceParameters", nuis)
        ws.defineSet("globalObservables", globs)
        ws.defineSet("Observables", obs)        
        ws.defineSet("POI", pois)
        mc.SetPdf(ws.pdf("CombinedPdf"))
        mc.SetObservables(ws.set("Observables"))
        mc.SetParametersOfInterest(ws.set("POI"))
        mc.SetNuisanceParameters(ws.set("nuisanceParameters"))
        mc.SetGlobalObservables(ws.set("globalObservables"))
        self.import_object(ws, mc, silent=False)
                    
    def generate_workspace(self):
        t0 = time.time()
        channel_list = ROOT.RooCategory("channellist", "channellist")
        combined_pdf = ROOT.RooSimultaneous("CombinedPdf", "", channel_list)
        
        obs          = ROOT.RooArgSet()
        pois         = ROOT.RooArgSet()
        nuis         = ROOT.RooArgSet()
        globs        = ROOT.RooArgSet()
        constraints  = ROOT.RooArgSet()
        
        self.reset()
        
        categories = list(self.core_config['input_categories'])
        
        for category in categories:
            ws_cat = self.generate_single_category(category)
            self.cat_ws[category] = ws_cat
            
            # import channel pdf
            channel_list.defineType(category)
            pdf_name = self.get_pdf_name(category)
            pdf_cat = ws_cat.pdf(pdf_name)
            if not pdf_cat:
                raise RuntimeError(f"model pdf `{pdf_name}` for the category "
                                   f"`{category}` is not defined")
            combined_pdf.addPdf(pdf_cat, category)
            
            obs.add(ws_cat.set(OBS_SET), True)
            pois.add(ws_cat.set(POI_SET), True)
            nuis.add(ws_cat.set(NUIS_SET), True)
            globs.add(ws_cat.set(GLOB_SET), True)
            if self.debug_mode:
                ws_i.Print()
                
        self.stdout.info("REGTEST: Making combined workspace")
                
        # implement main pdf
        self.import_object(self.comb_ws, combined_pdf)
        
        # implement model config
        self.implement_model_config_and_sets(self.comb_ws, self.model_config,
                                             obs, pois, nuis, globs)
        
        weight_var = ROOT.RooRealVar("wt", "wt", 1)
        args = ROOT.RooArgSet()
        args.add(obs)
        args.add(weight_var)
        
        dataset_map = ExtendedModel.get_dataset_map(self.datasets)
        dataset_binned_map = ExtendedModel.get_dataset_map(self.datasets_binned)
        
        dataset_name = self.core_config['dataset_name']
        obs_data = ROOT.RooDataSet(dataset_name, "Combined data", args,
                                   ROOT.RooFit.Index(channel_list),
                                   ROOT.RooFit.Import(dataset_map),
                                   ROOT.RooFit.WeightVar(weight_var))
        obs_data_binned = ROOT.RooDataSet(dataset_name + "binned", 
                                          "Binned combined data", args,
                                          ROOT.RooFit.Index(channel_list),
                                          ROOT.RooFit.Import(dataset_binned_map),
                                          ROOT.RooFit.WeightVar(weight_var))
        self.import_object(self.comb_ws, obs_data, silent=False)
        if (obs_data_binned.numEntries() < obs_data.numEntries()):
            self.import_object(self.comb_ws, obs_data_binned, silent=False)
        else:
            self.stdout.warning("REGTEST: No need to keep binned dataset, as the "
                                "number of data events is smaller than or equal "
                                "to the number of bins in all categories", "red")
            self.use_binned = False
        
        # save the original snapshot
        # self.comb_ws.saveSnapshot("nominalNuis", self.model_config.GetNuisanceParameters())
        # self.comb_ws.saveSnapshot("nominalGlobs", self.model_config.GetGlobalObservables())
        
        # generate asimov datasets
        self.generate_asimov()
        
        # import class code
        class_code_dirs = self.core_config["class_code_dirs"]
        if (class_code_dirs["decl"] is not None) and (class_code_dirs["impl"] is not None):
            self.stdout.warning("REGTEST: Importing class code")            
            self.comb_ws.addClassDeclImportDir(class_code_dirs["decl"])
            self.comb_ws.addClassImplImportDir(class_code_dirs["impl"])
            self.comb_ws.importClassCode()
            
        # save final workspace
        self.path_manager.makedir_for_files("output")
        output_path = self.path_manager.get_file("output")
        self.comb_ws.writeToFile(output_path, True)
        
        t1 = time.time()
        self.time_taken += (t1 - t0)
        self.print_final_summary()


from typing import Union, Optional, Dict, List, Set
import re
import uuid
import fnmatch

import ROOT

from .string_utils import remove_whitespace

def copy_attributes(source:"ROOT.RooAbsArg", target:"ROOT.RooAbsArg"):
    if source is target:
        return
    for attrib in source.attributes():
        target.setAttribute(attrib)
    for attrib in source.stringAttributes():
        target.setStringAttribute(attrib.first, attrib.second)
        
def get_variable_attributes(variable:ROOT.RooAbsReal, asym_error:bool=False):
    if isinstance(variable, ROOT.RooRealVar):
        if asym_error:
            attributes = {
                'name' : variable.GetName(),
                'value': variable.getVal(),
                'errorlo': variable.getErrorLo(),
                'errorhi': variable.getErrorHi(),
                'min'  : variable.getMin(),
                'max'  : variable.getMax(),
                'is_constant': variable.isConstant()
            }
        else:
            attributes = {
                'name' : variable.GetName(),
                'value': variable.getVal(),
                'error': variable.getError(),
                'min'  : variable.getMin(),
                'max'  : variable.getMax(),
                'is_constant': variable.isConstant()
            }
    elif isinstance(variable, ROOT.RooAbsReal):
        attributes = {
            'name' : variable.GetName(),
            'value': variable.getVal()
        }
    else:
        raise ValueError("variable must be an instance of RooAbsReal")
    return attributes

def variable_collection_to_dataframe(var_collection:Union[list, ROOT.RooArgSet],
                                     asym_error:bool=False):
    import pandas as pd
    data = []
    for variable in var_collection:
        attributes = get_variable_attributes(variable, asym_error=asym_error)
        data.append(attributes)
    df = pd.DataFrame(data)
    return df
        
def construct_categorized_pdf_dataset(pdf:"ROOT.RooAbsPdf", dataset:"ROOT.RooDataSet", 
                                      workspace:"ROOT.RooWorkspace", label:str,
                                      category_name:str="adhocCat"):
    # make sure pdf is not already a simultaneous pdf
    assert pdf.ClassName() != "RooSimultaneous"
    cat = ROOT.RooCategory(category_name, category_name)
    cat.defineType(label)
    pdf_dict = {label: pdf}
    pdf_map = ExtendedModel.get_object_map(pdf_dict, "RooAbsPdf")
    dataset_dict = {label: dataset}
    dataset_map = ExtendedModel.get_object_map(data_dict, "RooDataSet")
    sim_pdf = ROOT.RooSimultaneous(pdf.GetName(), pdf.GetName(), pdf_map, cat)
    obs_and_weight = dataset.get()
    weight_var = workspace.var("weightVar")
    if not weitht_var:
        raise RuntimeError("workspace does not contain the variable `weightVar`")
    obs_and_weight.add(weight_var)
    indexed_dataset = ROOT.RooDataSet(dataset.GetName(), dataset.GetName(), obs_and_weight, 
                                      ROOT.RooFit.Index(cat),
                                      ROOT.RooFit.Import(dataset_map),
                                      ROOT.RooFit.WeightVar(weight_var))
    return sim_pdf, indexed_dataset


def factorize_pdf(observables:"ROOT.RooArgSet", pdf:"ROOT.RooAbsPdf", constraints:"ROOT.RooArgSet"):
    pdf_class = pdf.ClassName()   
    if pdf_class == "RooProdPdf":
        new_factors = ROOT.RooArgList()
        new_owned = ROOT.RooArgSet()
        pdf_list = pdf.pdfList()
        need_new = False
        for i in range(len(pdf_list)):
            pdf_i = pdf_list[i]
            new_pdf = factorize_pdf(observables, pdf_i, constraints)
            if new_pdf == 0:
                need_new = True
                continue
            if new_pdf is not pdf_i:
                need_new = True
                new_owned.add(new_pdf)
            new_factors.add(new_pdf)
        if not need_new:
            return pdf
        elif len(new_factors) == 0:
            return 0
        elif len(new_factors) == 1:
            clone_pdf = new_factors.first().Clone("{}_obsOnly".format(pdf.GetName()))
            copy_attributes(pdf, clone_pdf)
            return clone_pdf
        factorized_pdf = ROOT.RooProdPdf("{}_obsOnly".format(pdf.GetName()), "", new_factors)
        factorized_pdf.addOwnedComponents(new_owned)
        copy_attributes(pdf, factorized_pdf)
        return factorized_pdf
    elif pdf_class == "RooSimultaneous":
        cat = pdf.indexCat().Clone()
        n_bins = cat.numBins("")
        factorized_pdfs = []
        new_owned = ROOT.RooArgSet()
        need_new = False
        for i in range(n_bins):
            cat.setBin(i)
            pdf_i = pdf.getPdf(cat.getLabel())
            new_pdf = factorize_pdf(observables, pdf_i, constraints)
            factorized_pdfs.append(new_pdf)     
            if new_pdf == 0:
                raise RuntimeError("channel `{}` factorized to 0".format(cat.getLabel()))
            if new_pdf is not pdf_i:
                need_new = True
                new_owned.add(new_pdf)
                # this can be removed after version 6.28
                ROOT.SetOwnership(new_pdf, False)
        factorized_pdf = pdf
        if need_new:
            factorized_pdf = ROOT.RooSimultaneous("{}_obsOnly".format(pdf.GetName()), "", pdf.indexCat())
            for i in range(n_bins):
                cat.setBin(i)
                new_pdf = factorized_pdfs[i]
                if new_pdf:
                    factorized_pdf.addPdf(new_pdf, cat.getLabel())
            factorized_pdf.addOwnedComponents(new_owned)
        # has to delete persistent object
        cat.Delete()
        copy_attributes(pdf, factorized_pdf)
        return factorized_pdf         
    elif pdf.dependsOn(observables):
        return pdf
    else:
        if not constraints.contains(pdf):
            constraints.add(pdf)
        return 0
    
    
def rebuild_simultaneous_pdf(observables:"ROOT.RooArgSet", sim_pdf:"ROOT.RooSimultaneous"):
    assert sim_pdf.ClassName() == "RooSimultaneous"
    constraints = ROOT.RooArgList()
    cat = sim_pdf.indexCat().Clone()
    n_bins = cat.numBins("")
    factorized_pdfs = []
    new_owned = ROOT.RooArgSet()
    for i in range(n_bins):
        cat.setBin(i)
        pdf_i = sim_pdf.getPdf(cat.getLabel())
        if pdf_i == 0:
            factorized_pdfs.append(0)
            continue
        new_pdf = factorize_pdf(observables, pdf_i, constraints)     
        factorized_pdfs.append(new_pdf)
        if new_pdf == 0:
            continue
        if new_pdf is not pdf_i:
            new_owned.add(new_pdf)
            # this can be removed after version 6.28
            ROOT.SetOwnership(new_pdf, False)
    rebuilt_pdf = ROOT.RooSimultaneous("{}_reloaded".format(sim_pdf.GetName()), "", sim_pdf.indexCat())
    for i in range(n_bins):
        cat.setBin(i)
        new_pdf = factorized_pdfs[i]
        if new_pdf:
            if constraints.getSize() > 0:
                all_factors = ROOT.RooArgList(constraints)
                all_factors.add(new_pdf)
                newer_pdf = ROOT.RooProdPdf("{}_plus_constr".format(new_pdf.GetName()), "",
                                           all_factors)
                rebuilt_pdf.addPdf(newer_pdf, cat.getLabel())
                copy_attributes(new_pdf, newer_pdf)
                new_owned.add(newer_pdf)
                # this can be removed after version 6.28
                ROOT.SetOwnership(newer_pdf, False)
            else:
                rebuilt_pdf.addPdf(new_pdf, cat.getLabel())
    rebuilt_pdf.addOwnedComponents(new_owned)
    copy_attributes(sim_pdf, rebuilt_pdf)
    return rebuilt_pdf

def print_object(obj, indent=0, spacer="  ", prefix="-", show_address=False):
    if show_address:
        print(f"{spacer*indent}{prefix}{obj.GetName()}({obj.ClassName()} @ {hex(id(obj))})")
    else:
        print(f"{spacer*indent}{prefix}{obj.GetName()}({obj.ClassName()})")
        
def print_pdf_structure(pdf:"ROOT.RooAbsPdf", level:int=0, show_address=False, max_level:int=-1):
    if (max_level >= 0) and (level > max_level):
        return None
    print_object(pdf, level, show_address=show_address)
    class_name = pdf.ClassName()
    if class_name == "RooSimultaneous":
        cat = pdf.indexCat().Clone()
        n_bins = cat.numBins("")
        for i in range(n_bins):
            cat.setBin(i)
            pdf_i = pdf.getPdf(cat.getLabel())
            print_pdf_structure(pdf_i, level+1, show_address=show_address, max_level=max_level)
    elif class_name == "RooProdPdf":
        pdf_list = pdf.pdfList()
        for pdf_i in pdf_list:
            print_pdf_structure(pdf_i, level+1, show_address=show_address, max_level=max_level)
    elif class_name == "RooRealSumPdf":
        pdf_list = pdf.getComponents()
        for pdf_i in pdf_list:
            if pdf_i == pdf:
                continue
            print_pdf_structure(pdf_i, level+1, show_address=show_address, max_level=max_level)
                     
def get_correlation_matrix(roofit_result, lib:str="pandas"):
    correlation_hist = roofit_result.correlationHist()
    if lib.lower() == "pandas":
        from quickstats.interface.root import TH2
        df = TH2.correlationHist_to_dataframe(correlation_hist)
        correlation_hist.Delete()
        return df
    elif lib.lower() == "root":
        return correlation_hist
    else:
        raise ValueError(f"unsupported library: {lib} (available options: \"root\", \"pandas\")")
        
def get_relevant_components(target, reference:Set[ROOT.RooArgSet]):
    result = ROOT.RooArgSet()
    if target in reference:
        result.add(target)
    components = None
    if isinstance(target, ROOT.RooAbsPdf):
        components = target.getComponents()
        components.remove(target)
    elif isinstance(target, ROOT.RooProduct):
        components = target.components()
    if (components is not None) and components.getSize():
        for component in components:
            relevant_components = get_relevant_components(component, reference)
            result.add(relevant_components)
    return result

def export_hist_data(data:ROOT.RooDataSet,
                     pdf:Optional[ROOT.RooAbsPdf]=None,
                     observable:Optional[ROOT.RooRealVar]=None,
                     bin_range:Optional[List]=None,
                     bin_error:bool=True,
                     n_bins_data:Optional[int]=None,
                     n_bins_pdf:Optional[int]=1000):

    if observable is None:
        if data.get().size() > 1:
            raise RuntimeError("only single-observable is allowed")
        observable = data.get().first()

    if bin_range is None:
        bin_range = observable.getRange()
        bin_range = [bin_range.first, bin_range.second]

    if n_bins_data is None:
        n_bins_data = observable.getBins()
        
    if n_bins_pdf is None:
        n_bins_pdf = n_bins_data        
        
    binning_data = ROOT.RooFit.Binning(n_bins_data, bin_range[0], bin_range[1])
    binning_pdf  = ROOT.RooFit.Binning(n_bins_pdf, bin_range[0], bin_range[1])

    from quickstats.interface.root import RooAbsData
    h_data = RooAbsData.create_histogram(data, uuid.uuid4().hex, observable, binning_data)
    
    if bin_error:
        # use SumW2 error if weighted, else use Poisson error
        if data.isWeighted():
            bin_errors = {"lo": h_data.bin_error, "hi": h_data.bin_error}
        else:
            from quickstats.interface.root import TH1
            bin_errors = TH1.GetPoissonError(h_data.bin_content)
    else:
        size = len(h_data.bin_content)
        bin_errors = {"lo": np.zeros(size), "hi": np.zeros(size)}

    hist_data = {
        'data': {
            'x': h_data.bin_center,
            'y': h_data.bin_content,
            'yerrlo': bin_errors['lo'],
            'yerrhi': bin_errors['hi']
        }
    }
    
    if pdf is not None:
        
        h_pdf = RooAbsData.create_histogram(pdf, uuid.uuid4().hex, observable, binning_pdf)
        h_pdf_data_binning = RooAbsData.create_histogram(pdf, uuid.uuid4().hex, observable,
                                                         binning_data)
        
        # calculate normalization factors
        norm_data  = h_data.bin_content.sum()
        norm_pdf1  = h_pdf.bin_content.sum()
        norm_pdf2  = h_pdf_data_binning.bin_content.sum()            
            
        hist_data['pdf'] = {
            'x': h_pdf.bin_center,
            'y': h_pdf.bin_content * norm_data / norm_pdf1 * (n_bins_pdf / n_bins_data)
        }
        hist_data['pdf_data_binning'] = {
            'x': h_pdf_data_binning.bin_center,
            'y': h_pdf_data_binning.bin_content * norm_data / norm_pdf2
        }
        
    return hist_data

def translate_formula(name:str, formula:str):
    variable_regex = re.compile(r"\b[a-zA-Z]\w+\b")
    variables = variable_regex.findall(formula)
    variables = sorted(set(variables))
    expr = formula
    for i, variable in enumerate(variables):
        expr = re.sub(r"\b" + variable + r"\b", f"@{i}", expr)
    translated_formula = f"expr::{name}('{expr}', {', '.join(variables)})"
    return translated_formula, variables

def recover_formula(formula:str):
    expr_regex = re.compile(r"'(.*?)'")
    expressions = expr_regex.findall(formula)
    if len(expressions) != 1:
        raise RuntimeError("invalid RooFit factory expression")
    expression = expressions[0]
    variables_regex = re.compile(r"\('.*?',(.*)\)")
    variables = variables_regex.findall(formula)
    if len(variables) != 1:
        raise RuntimeError("invalid RooFit factory expression")
    variables = [i.strip() for i in variables[0].split(",") if i.strip()]
    recovered_formula = expression
    for i, variable in enumerate(variables):
        recovered_formula = re.sub(f"@{i}(?![0-9])", variable, recovered_formula)
    return recovered_formula

def get_nuis_beta_terms(nuis_name:str, components):
    nuis_term = None
    beta_term = None
    if components.size() != 2:
        return nuis_term, beta_term
    for component in components:
        if component.GetName() == nuis_name:
            nuis_term = component
        else:
            beta_term = component
    return nuis_term, beta_term

def get_gaus_response_variations(nuis:ROOT.RooRealVar, client:ROOT.RooAddition):
    result = {"nominal": None, "low": None, "high": None, "type": None}
    if not isinstance(client, ROOT.RooAddition):
        raise ValueError("gaussian response function must be an instance of RooAddition")
    nuis_name = nuis.GetName()
    nominal_term, beta_term, nuis_term, resp_term = None, None, None
    terms = client.list()
    for term in terms:
        if isinstance(term, ROOT.RooProduct):
            for subterm in term.components():
                if isinstance(subterm, ROOT.RooProduct):
                    nuis_term, beta_term = get_nuis_beta_terms(nuis_name, subterm.components())
                elif isinstance(subterm , ROOT.RooRealVar):
                    resp_term = subterm
        elif isinstance(term, ROOT.RooRealVar):
            nominal_term = term
    if any(term is None for term in [beta_term, nuis_term, nominal_term, resp_term]):
        return result
    nominal = nominal_term.getVal()
    magnitude = resp_term.getVal()
    beta = beta_term.getVal()
    value = round(magnitude * beta, 8)
    return {"nominal": nominal, "low": value, "high": value, "type": "gaus"}

def get_logn_response_variations(nuis:ROOT.RooRealVar, client:ROOT.RooFormulaVar):
    result = {"nominal": None, "low": None, "high": None, "type": None}
    if not isinstance(client, ROOT.RooFormulaVar):
        raise ValueError("lognormal response function must be an instance of RooFormulaVar")    
    nuis_name = nuis.GetName()
    formula = client.formula()
    formula_str = formula.formulaString()
    formula_str = remove_whitespace(formula_str)
    if not formula_str.startswith("exp("):
        return result
    dependents = formula.actualDependents()
    if dependents.size() != 2:
        return result
    beta_term, nuis_term, resp_term = None, None, None
    for dependent in dependents:
        if isinstance(dependent, ROOT.RooProduct):
            nuis_term, beta_term = get_nuis_beta_terms(nuis_name, dependent.components())
        elif isinstance(dependent, ROOT.RooFormulaVar):
            resp_term = dependent
    if any(term is None for term in [beta_term, nuis_term, resp_term]):
        return result
    beta = beta_term.getVal()
    resp_formula_str = resp_term.formula().formulaString()
    resp_formula_str = remove_whitespace(resp_formula_str)
    if resp_formula_str == "log(1+x[0]/x[1])":
        resp_dependents = resp_term.formula().actualDependents()
        magnitude = resp_dependents[0].getVal()
        value = round(magnitude * beta, 8)
        nominal = resp_dependents[1].getVal()
        return {"nominal": nominal, "low": value, "high": value, "type": "logn"}
    return result

def get_asym_response_variations(nuis:ROOT.RooRealVar, client):
    if not isinstance(client, ROOT.RooStats.HistFactory.FlexibleInterpVar):
        raise ValueError("asymmetric response function must be an instance of RooFormulaVar") 
    low = client.low()
    high = client.high()
    nominal = client.nominal()
    if (low.size() != 1) or (high.size() != 1):
        raise RuntimeError(f'invalid low/high value format for the FlexibleInterpVar '
                           f'intance "{client.GetName()}"')
    return {"nominal": nominal, "low": round(low[0], 8), "high": round(high[0], 8), "type": "asym"}

def get_ProcessNormalization_response_variations(nuis:ROOT.RooRealVar, client):
    if client.ClassName() != "ProcessNormalization":
        raise ValueError("response function must be an instance of ProcessNormalization")
    nuis_name = nuis.GetName()
    kappas = client.getKappa(nuis_name)
    if not kappas:
        raise RuntimeError(f'failed to extract magnitude for the systematic "{nuis_name}" from '
                           f'the ProcessNormalization client "{client.GetName()}"')
    if len(kappas) == 1:
        return {"nominal": 1, "low":  round(kappas[0] - 1, 8), "high": round(kappas[0] - 1, 8), "type": "logn"}
    elif len(kappas) == 2:
        return {"nominal": 1, "low":  round(kappas[0] - 1, 8), "high": round(kappas[1] - 1, 8), "type": "asym"}
    else:
        raise RuntimeError(f'invalid kappa values for the systematic "{nuis_name}" from '
                           f'the ProcessNormalization client "{client.GetName()}"')
            
def get_systematic_variations_from_client(nuis:ROOT.RooRealVar, client):
    result = {"client": None, "nominal": None, "low": None, "high": None, "type": None}
    class_name = client.ClassName()
    nuis_name = nuis.GetName()
    # CMS dedicated normalization response function
    if class_name == "ProcessNormalization":
        result["client"] = client.GetName()
        result.update(get_ProcessNormalization_response_variations(nuis, client))
    elif class_name == "RooProduct":
        parent_clients = client.clients()
        if parent_clients.size() != 1:
            return result
        parent_client = parent_clients[0]
        result["client"] = parent_client.GetName()
        parent_class_name = parent_client.ClassName()
        if isinstance(parent_client, ROOT.RooFormulaVar):
            result.update(get_logn_response_variations(nuis, parent_client))
        elif isinstance(parent_client, ROOT.RooAddition):
            result.update(get_gaus_response_variations(nuis, parent_client))
        elif isinstance(parent_client, ROOT.RooStats.HistFactory.FlexibleInterpVar):
            result.update(get_asym_response_variations(nuis, parent_client))
    return result

def get_systematic_variations_from_clients(nuis:ROOT.RooRealVar, constr_pdf:ROOT.RooAbsPdf):
    constr_pdf_name = constr_pdf.GetName()
    clients = [c for c in nuis.clients() if c.GetName() != constr_pdf_name]
    result = {}
    for client in clients:
        client_result = get_systematic_variations_from_client(nuis, client)
        client_name = client_result.pop("client")
        if client_name is not None:
            result[client_name] = client_result
    return result

def get_systematics_variations(nuis_set:Union[List, ROOT.RooArgSet], constr_set:Union[List, ROOT.RooArgSet],
                               filter_name:Optional[str]=None,
                               filter_client:Optional[str]=None,
                               fmt:str="pandas"):
    assert len(nuis_set) == len(constr_set)
    result = {}
    for nuis, constr in zip(nuis_set, constr_set):
        nuis_name = nuis.GetName()
        result[nuis_name] = get_systematic_variations_from_clients(nuis, constr)
    import pandas as pd
    data = []
    for nuis_name, client_results in result.items():
        for client_name, client_result in client_results.items():
            data.append({"name": nuis_name, "client": client_name, **client_result})
    df = pd.DataFrame(data)
    if filter_name is not None:
        df = df[df.apply(lambda x: fnmatch.fnmatch(x["name"], filter_name), axis=1)]
    if filter_client is not None:
        df = df[df.apply(lambda x: fnmatch.fnmatch(x["client"], filter_client), axis=1)]
    if fmt.lower() == "dict":
        df = df.set_index(["name", "client"])
        return {level: df.xs(level).to_dict('index') for level in df.index.levels[0]}
    if fmt.lower() == "pandas":
        return df
    raise ValueError(f'invalid output format "{fmt}" (choose between "dict" or "pandas")')
    
def decompose_nll(nll:"ROOT.RooRealSum", global_observables:"ROOT.RooArgSet", fmt:str="pandas"):
    nll_main_components = [component for component in nll.list()]
    data_terms = [component for component in nll_main_components if component.ClassName() == "RooNLLVar"]
    assert len(data_terms) > 0
    constr_terms = [component for component in nll_main_components if component.ClassName() == "RooConstraintSum"]
    result = {
        "name": [],
        "class": [],
        "type": [],
        "nll":[]
    }
    for data_term in data_terms:
        result["name"].append(data_term.GetName())
        result["class"].append(data_term.ClassName())
        result["type"].append("data")
        result["nll"].append(data_term.getVal())
    for constr_term in constr_terms:
        for constr_pdf in constr_term.list():
            result["name"].append(constr_pdf.GetName())
            result["class"].append(constr_pdf.ClassName())
            result["type"].append("constraint")
            result["nll"].append(-constr_pdf.getLogVal(global_observables))
    if fmt.lower() == "dict":
        return result
    if fmt.lower() == "pandas":
        import pandas as pd
        return pd.DataFrame(result)
    raise ValueError(f'invalid output format "{fmt}" (choose between "dict" or "pandas")')
    
def get_pdf_attributes(pdfs:"ROOT.RooArgSet", attributes, pdf_class=None, fmt:str="pandas"):
    import quickstats
    if not (quickstats.root_version >= (6, 26, 0)):
        raise RuntimeError("this method requires ROOT version > 6.26.0")
    results = []
    for pdf in pdfs:
        if (pdf_class is not None) and (not isinstance(pdf, pdf_class)):
            continue
        result = {}
        result["name"] = pdf.GetName()
        result["value"] = pdf.getVal()
        for attribute in attributes:
            getter_method_str = f"get{attribute}"
            if not hasattr(pdf, getter_method_str):
                raise ValueError(f"pdf {pdf.GetName()}(class = {pdf.Class()}) does not contain the method {getter_method_str}")
            variable = getattr(pdf, getter_method_str)()
            result[f"{attribute.lower()} (name)"] = variable.GetName()
            result[f"{attribute.lower()} (value)"] = variable.getVal()
        results.append(result)
    if fmt.lower == "dict":
        return results
    if fmt.lower() == "pandas":
        import pandas as pd
        return pd.DataFrame(results)
    raise ValueError(f'invalid output format "{fmt}" (choose between "dict" or "pandas")')
    
def get_gaussian_pdf_attributes(pdfs:"ROOT.RooArgSet", fmt:str="pandas"):
    return get_pdf_attributes(pdfs, ["X", "Mean", "Sigma"], ROOT.RooGaussian, fmt=fmt)

def get_poisson_pdf_attributes(pdfs:"ROOT.RooArgSet", fmt:str="pandas"):
    return get_pdf_attributes(pdfs, ["X", "Mean"], pdf_class=ROOT.RooPoisson, fmt=fmt)
    


from typing import Dict, Optional, Union, List
import pandas as pd
import numpy as np

from quickstats.plots import General1DPlot
from quickstats.plots.template import create_transform
from quickstats.utils.common_utils import combine_dict

class Likelihood1DPlot(General1DPlot):
    
    STYLES = {
        'annotation':{
            'fontsize': 20
        }    
    }    
    
    CONFIG = {
        'sigma_values': (1, 4),
        'sigma_pos': 0.93,
        'sigma_names': ('1 $\sigma$', '2 $\sigma$'),
        'sigma_line_styles':{
            'color': 'gray',
            'linestyle': '--'
        },
        'sigma_interval_styles':{
            'loc': (0.2, 0.4),
            'main_text': '',
            'sigma_text': r'{sigma_name}: {xlabel}$\in {intervals}$',
            'dy': 0.05,
            'decimal_place': 2
        }
    }
    
    def __init__(self, data_map:Union[pd.DataFrame, Dict[str, pd.DataFrame]],
                 label_map:Optional[Dict]=None,
                 styles_map:Optional[Dict]=None,
                 color_cycle=None,
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Dict]=None,
                 config:Optional[Dict]=None):
        super().__init__(data_map=data_map,
                         label_map=label_map,
                         styles_map=styles_map,
                         color_cycle=color_cycle,
                         styles=styles,
                         analysis_label_options=analysis_label_options,
                         config=config)
        self.intersections = {}
    
    def draw_sigma_intervals(self, ax, x:np.ndarray, y:np.ndarray, xlabel:str=""):
        from quickstats.maths.interpolation import get_intersections
        sigma_values = self.config['sigma_values']
        sigma_names  = self.config['sigma_names']
        anno_loc = self.config['sigma_interval_styles']['loc']
        anno_main_text = self.config['sigma_interval_styles']['main_text']
        anno_sigma_text = self.config['sigma_interval_styles']['sigma_text']
        anno_dy = self.config['sigma_interval_styles']['dy']
        dp = self.config['sigma_interval_styles']['decimal_place']
        ax.annotate(anno_main_text, anno_loc, xycoords='axes fraction', **self.styles['annotation'])
        for i, (sigma_value, sigma_name) in enumerate(zip(sigma_values, sigma_names)):
            intersections = get_intersections(x, y, sigma_value)
            self.intersections[sigma_name] = intersections
            intersection_str = r" \cup ".join([f"[{lo:.{dp}f}, {hi:.{dp}f}]" for (lo, hi) in intersections])
            sigma_text = anno_sigma_text.format(sigma_name=sigma_name, xlabel=xlabel, intervals=intersection_str)
            ax.annotate(sigma_text, (anno_loc[0], anno_loc[1] - (i + 1) * anno_dy),
                        xycoords='axes fraction', **self.styles['annotation'])

    def draw(self, xattrib:str='mu', yattrib:str='qmu', xlabel:Optional[str]=None, 
             ylabel:Optional[str]="$-2\Delta ln(L)$",
             ymin:float=0, ymax:float=7, xmin:Optional[float]=None, xmax:Optional[float]=None,
             draw_sigma_line:bool=True, draw_sm_line:bool=False,
             draw_sigma_intervals:Union[str, bool]=False):
                        
        ax = super().draw(xattrib=xattrib, yattrib=yattrib,
                          xlabel=xlabel, ylabel=ylabel,
                          ymin=ymin, ymax=ymax, xmin=xmin, xmax=xmax)

        if draw_sigma_line:
            transform = create_transform(transform_x="axis", transform_y="data")
            sigma_line_styles = self.config['sigma_line_styles']
            sigma_values = self.config['sigma_values']
            sigma_names = self.config['sigma_names']
            ax.hlines(sigma_values, xmin=0, xmax=1, zorder=0, transform=transform,
                      **sigma_line_styles)
            if sigma_names:
                sigma_pos = self.config['sigma_pos']
                for sigma_value, sigma_name in zip(sigma_values, sigma_names):
                    ax.text(sigma_pos, sigma_value, sigma_name, color='gray', ha='left',
                            va='bottom' if (sigma_pos > 0 and sigma_pos < 1) else 'center', fontsize=20, transform=transform)
            
        if draw_sm_line:
            transform = create_transform(transform_y="axis", transform_x="data")
            sm_line_styles = self.config['sm_line_styles']
            sm_values = self.config['sm_values']
            sm_names = self.config['sm_names']
            ax.vlines(sm_values, ymin=0, ymax=1, zorder=0, transform=transform,
                      **sm_line_styles)
            if sm_names:
                sm_pos = self.config['sm_pos']
                for sm_value, sm_name in zip(sm_values, sm_names):
                    ax.text(sm_value, sm_pos, sm_name, color='gray', ha='right', rotation=90,
                            va='bottom' if (sm_pos > 0 and sm_pos < 1) else 'center', fontsize=20, transform=transform)
        if draw_sigma_intervals:
            if isinstance(self.data_map, pd.DataFrame):
                x = self.data_map[xattrib]
                y = self.data_map[yattrib]
            elif isinstance(self.data_map, dict):
                if not isinstance(draw_sigma_intervals, str):
                    raise RuntimeError("name of the target likelihood curve must be specified "
                                       "when drawing sigma intervals")
                target = draw_sigma_intervals
                x = self.data_map[target][xattrib]
                y = self.data_map[target][yattrib]
            else:
                raise ValueError("invalid data format")
            self.draw_sigma_intervals(ax, x, y, xlabel=xlabel)
        return ax



from typing import Optional, Union, Dict, List

import pandas as pd
import numpy as np

from matplotlib import colors
from matplotlib.ticker import MaxNLocator
from matplotlib.lines import Line2D
from matplotlib.patches import Polygon

from quickstats.plots import AbstractPlot
from quickstats.plots.template import single_frame, parse_styles, create_transform
from quickstats.utils.common_utils import combine_dict
from quickstats.maths.statistics import min_max_to_range

class VariableDistributionPlot(AbstractPlot):
    
    def __init__(self, data_map:Dict[str, pd.DataFrame], plot_options:Dict[str, Dict],
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Dict]=None,
                 config:Optional[Dict]=None):
        """
        Arguments:
            plot_options: dicionary
                A dictionary containing plot options for various group of samples.
                Format: { <sample_group>: {
                            "samples": <list of sample names>,
                            "styles": <options in mpl.hist>},
                            "type": "hist" or "errorbar"
                          ...}
             
        """
        self.data_map = data_map
        self.plot_options = plot_options
        super().__init__(styles=styles,
                         analysis_label_options=analysis_label_options,
                         config=config)
    
    def draw(self, column_name:str, weight_name:Optional[str]="weight",
             xlabel:str="", ylabel:str="Fraction of Events / {bin_width:.2f}",
             nbins:int=25, xmin:Optional[float]=None, xmax:Optional[float]=None,
             ypad:float=0.4, rescale_by:Optional[float]=None,
             normalize:bool=True):
        """
        
        Arguments:
            column_name: string, default = "score"
                Name of the variable in the dataframe.
            weight_name: (optional) string, default = "weight"
                If specified, normalize the histogram by the "weight_name" variable
                in the dataframe.
            xlabel: string, default = "Score"
                Label of x-axis.
            ylabel: string, default = "Fraction of Events / {bin_width}"
                Label of y-axis.
            boundaries: (optional) list of float
                If specified, draw score boundaries at given values.
            nbins: int, default = 25
                Number of histogram bins.
            xmin: (optional) float
                Minimum value of the bins.
            xmax: (optional) float
                Maximum value of the bins.
            ypad: float, default = 0.4
                Fraction of the y-axis that should be padded.
            rescale_by: (optional) float
                Rescale variable values by a factor.
            normalize: bool, default = True
                Normalize histogram to unity.
        """
        ax = self.draw_frame()
        data_xmin = None
        data_xmax = None
        bin_range = min_max_to_range(xmin, xmax)
        for key in self.plot_options:
            samples = self.plot_options[key]["samples"]
            plot_style = self.plot_options[key].get("styles", {})
            plot_type = self.plot_options[key].get("type", "hist")
            df = pd.concat([self.data_map[sample] for sample in samples], ignore_index = True)
            x = df[column_name].values
            if rescale_by is not None:
                x = x * rescale_by
            if data_xmin is None:
                data_xmin = np.min(x)
            else:
                data_xmin = min(data_xmin, np.min(x))
            if data_xmax is None:
                data_xmax = np.max(x)
            else:
                data_xmax = max(data_xmax, np.max(x))
            if weight_name is not None:
                weights = df[weight_name]
            else:
                weights = None
            if plot_type == "hist":
                if normalize and (weights is not None):
                    weights = weights / weights.sum()
                plot_style = combine_dict(self.styles["hist"], plot_style)
                y, x, _ = ax.hist(x, nbins, range=bin_range,
                                  weights=weights, **plot_style, zorder=-5)
            elif plot_type == "errorbar":
                from quickstats.maths.statistics import get_hist_data
                hist_data = get_hist_data(x, weights,
                                          normalize=normalize,
                                          range=bin_range,
                                          bins=nbins)
                plot_style = combine_dict(self.styles["errorbar"], plot_style)
                ax.errorbar(**hist_data, **plot_style)
            else:
                raise RuntimeError(f'unknown plot type: {plot_type}')
        if bin_range is None:
            axis_xmin = data_xmin
            axis_xmax = data_xmax
        else:
            axis_xmin = bin_range[0]
            axis_xmax = bin_range[1]
        bin_width = (xmax - xmin) / nbins
        ylabel = ylabel.format(bin_width=bin_width)
        
        self.draw_axis_components(ax, xlabel=xlabel, ylabel=ylabel)
        self.set_axis_range(ax, xmin=axis_xmin, xmax=axis_xmax, ypad=ypad)

        handles, labels = ax.get_legend_handles_labels()
        new_handles = [Line2D([], [], c=h.get_edgecolor(), linestyle=h.get_linestyle(),
                       **self.styles['legend_Line2D']) if isinstance(h, Polygon) else h for h in handles]
        ax.legend(handles=new_handles, labels=labels, **self.styles['legend'])
        return ax



class TObject:
    @property
    def obj(self):
        return self._obj
    
    @obj.setter
    def obj(self, ref):
        if ref is None:
            self._obj = None
        else:
            fundamental_type = self.get_fundamental_type()
            if not isinstance(ref, fundamental_type):
                raise ValueError(f"Invalid ROOT object. Object must be an instance of {fundamental_type}.")
        self._obj = ref
    
    def __init__(self, obj=None):
        self.obj = obj
        
    def get_fundamental_type(self):
        raise NotImplementedError